#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import base64
import datetime
import functools
import json
import os
import platform
import re
import time
import urllib
import warnings
from typing import Optional, TypeVar
from urllib.parse import urlencode

import atatus
from atatus.base import Client, get_client
from atatus.conf import constants
from atatus.traces import execution_context
from atatus.utils import encoding, get_name_from_func, nested_key
from atatus.utils.disttracing import TraceParent
from atatus.utils.logging import get_logger

SERVERLESS_HTTP_REQUEST = ("api", "elb")

logger = get_logger("atatus.serverless")

COLD_START = True
INSTRUMENTED = False
REGISTER_PARTIAL_TRANSACTIONS = True

_AWSLambdaContextT = TypeVar("_AWSLambdaContextT")


def capture_serverless(func: Optional[callable] = None, **kwargs) -> callable:
    """
    Decorator for instrumenting AWS Lambda functions.

    Example usage:

        from atatus import capture_serverless

        @capture_serverless
        def handler(event, context):
            return {"statusCode": r.status_code, "body": "Success!"}

    Please note that when using the APM Layer and setting AWS_LAMBDA_EXEC_WRAPPER this is not required and
    the handler would be instrumented automatically.
    """
    if not func:
        # This allows for `@capture_serverless()` in addition to
        # `@capture_serverless` decorator usage
        return functools.partial(capture_serverless, **kwargs)

    if kwargs:
        warnings.warn(
            PendingDeprecationWarning(
                "Passing keyword arguments to capture_serverless will be deprecated in the next major release."
            )
        )

    name = kwargs.pop("name", None)

    kwargs = prep_kwargs(kwargs)

    global INSTRUMENTED
    client = get_client()
    if not client:
        client = Client(**kwargs)
    if not client.config.debug and client.config.instrument and client.config.enabled and not INSTRUMENTED:
        atatus.instrument()
        INSTRUMENTED = True

    @functools.wraps(func)
    def decorated(*args, **kwds):
        if len(args) >= 2:
            # Saving these for request context later
            event, context = args[0:2]
        else:
            event, context = {}, {}

        if not client.config.debug and client.config.instrument and client.config.enabled:
            with _lambda_transaction(func, name, client, event, context) as sls:
                sls.response = func(*args, **kwds)
                return sls.response
        else:
            return func(*args, **kwds)

    return decorated


def prep_kwargs(kwargs=None):
    if kwargs is None:
        kwargs = {}

    # Disable all background threads except for transport
    kwargs["metrics_interval"] = "0ms"
    kwargs["breakdown_metrics"] = False
    if "metrics_sets" not in kwargs and "ATATUS_METRICS_SETS" not in os.environ:
        # Allow users to override metrics sets
        kwargs["metrics_sets"] = []
    kwargs["central_config"] = False
    kwargs["cloud_provider"] = "none"
    kwargs["framework_name"] = "AWS Lambda"
    if "service_name" not in kwargs and "ATATUS_SERVICE_NAME" not in os.environ:
        kwargs["service_name"] = os.environ["AWS_LAMBDA_FUNCTION_NAME"]
    if "service_version" not in kwargs and "ATATUS_SERVICE_VERSION" not in os.environ:
        kwargs["service_version"] = os.environ.get("AWS_LAMBDA_FUNCTION_VERSION")

    return kwargs


def should_normalize_headers(event: dict) -> bool:
    """
    Helper to decide if we should normalize headers or not depending on the event

    Even if the documentation says that headers are lowercased it's not always the case for format version 1.0
    https://docs.aws.amazon.com/apigateway/latest/developerguide/http-api-develop-integrations-lambda.html
    """

    request_context = event.get("requestContext", {})
    return ("elb" in request_context or "requestId" in request_context) and "http" not in request_context


class _lambda_transaction(object):
    """
    Context manager for creating transactions around AWS Lambda functions.

    Begins and ends a single transaction, waiting for the transport to flush
    before releasing the context.
    """

    def __init__(
        self, func: callable, name: Optional[str], client: Client, event: dict, context: _AWSLambdaContextT
    ) -> None:
        self.func = func
        self.name = name or get_name_from_func(func)
        self.event = event
        self.context = context
        self.response = None
        self.client = client

    def __enter__(self):
        """
        Transaction setup
        """
        if not isinstance(self.event, dict):
            # When `event` is not a dict, it's likely the output of another AWS
            # service like Step Functions, and is unlikely to be standardized
            # in any way. We just have to rely on our defaults in this case.
            self.event = {}

        headers = self.event.get("headers") or {}
        if headers and should_normalize_headers(self.event):
            normalized_headers = {k.lower(): v for k, v in headers.items()}
        else:
            normalized_headers = headers
        trace_parent = TraceParent.from_headers(normalized_headers)

        global COLD_START
        cold_start = COLD_START
        COLD_START = False

        self.source = "other"
        transaction_type = "request"
        transaction_name = os.environ.get("AWS_LAMBDA_FUNCTION_NAME", self.name)

        self.httpmethod = (
            nested_key(self.event, "requestContext", "httpMethod")
            or nested_key(self.event, "requestContext", "http", "method")
            or nested_key(self.event, "httpMethod")
        )

        if self.httpmethod:  # http request
            if nested_key(self.event, "requestContext", "elb"):
                self.source = "elb"
                resource = "unknown route"
            elif nested_key(self.event, "requestContext", "httpMethod"):
                self.source = "api"
                # API v1
                resource = "/{}{}".format(
                    nested_key(self.event, "requestContext", "stage"),
                    nested_key(self.event, "requestContext", "resourcePath"),
                )
            else:
                self.source = "api"
                # API v2
                route_key = nested_key(self.event, "requestContext", "routeKey")
                route_key = f"/{route_key}" if route_key.startswith("$") else route_key.split(" ", 1)[-1]
                resource = "/{}{}".format(
                    nested_key(self.event, "requestContext", "stage"),
                    route_key,
                )
            transaction_name = "{} {}".format(self.httpmethod, resource)
        elif "Records" in self.event and len(self.event["Records"]) == 1:
            record = self.event["Records"][0]
            if record.get("eventSource") == "aws:s3":  # S3
                self.source = "s3"
                transaction_name = "{} {}".format(record["eventName"], record["s3"]["bucket"]["name"])
            elif record.get("EventSource") == "aws:sns":  # SNS
                self.source = "sns"
                transaction_type = "messaging"
                transaction_name = "RECEIVE {}".format(record["Sns"]["TopicArn"].split(":")[5])
            elif record.get("eventSource") == "aws:sqs":  # SQS
                self.source = "sqs"
                transaction_type = "messaging"
                transaction_name = "RECEIVE {}".format(record["eventSourceARN"].split(":")[5])

        if "Records" in self.event:
            links = [
                TraceParent.from_string(record["messageAttributes"]["traceparent"]["stringValue"])
                for record in self.event["Records"][:1000]
                if "messageAttributes" in record and "traceparent" in record["messageAttributes"]
            ]
        else:
            links = []

        self.client.begin_transaction(transaction_type, trace_parent=trace_parent, links=links)
        atatus.set_transaction_name(transaction_name, override=False)
        if self.source in SERVERLESS_HTTP_REQUEST:
            atatus.set_context(
                lambda: get_data_from_request(
                    self.event,
                    capture_body=self.client.config.capture_body in ("request", "all"),
                    capture_headers=self.client.config.capture_headers,
                ),
                "request",
            )
        self.set_metadata_and_context(cold_start)
        self.send_partial_transaction()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """
        Transaction teardown
        """
        if self.response and isinstance(self.response, dict):
            atatus.set_context(
                lambda: get_data_from_response(self.response, capture_headers=self.client.config.capture_headers),
                "response",
            )
            if "statusCode" in self.response:
                try:
                    result = "HTTP {}xx".format(int(self.response["statusCode"]) // 100)
                    atatus.set_transaction_result(result, override=False)
                    if result == "HTTP 5xx":
                        atatus.set_transaction_outcome(outcome="failure", override=False)
                except ValueError:
                    logger.warning("Lambda function's statusCode was not formed as an int. Assuming 5xx result.")
                    atatus.set_transaction_result("HTTP 5xx", override=False)
                    atatus.set_transaction_outcome(outcome="failure", override=False)
        if exc_val:
            self.client.capture_exception(exc_info=(exc_type, exc_val, exc_tb), handled=False)
            if self.source in SERVERLESS_HTTP_REQUEST:
                atatus.set_transaction_result("HTTP 5xx", override=False)
                atatus.set_transaction_outcome(http_status_code=500, override=False)
                atatus.set_context({"status_code": 500}, "response")
            else:
                atatus.set_transaction_result("failure", override=False)
                atatus.set_transaction_outcome(outcome="failure", override=False)

        self.client.end_transaction()
        # Collect any custom+prometheus metrics if enabled
        self.client.metrics.collect()

        try:
            logger.debug("Flushing atatus data")
            self.client._transport.flush()
            logger.debug("Flush complete")
        except ValueError:
            logger.warning("Flush timed out")

    def set_metadata_and_context(self, coldstart: bool) -> None:
        """
        Process the metadata and context fields for this request
        """
        metadata = {}
        cloud_context = {"origin": {"provider": "aws"}}
        service_context = {}
        message_context = {}

        faas = {}
        faas["coldstart"] = coldstart
        faas["trigger"] = {"type": "other"}
        faas["execution"] = self.context.aws_request_id
        arn = self.context.invoked_function_arn
        if len(arn.split(":")) > 7:
            arn = ":".join(arn.split(":")[:7])
        faas["id"] = arn
        faas["name"] = os.environ.get("AWS_LAMBDA_FUNCTION_NAME")
        faas["version"] = os.environ.get("AWS_LAMBDA_FUNCTION_VERSION")

        if self.source == "api":
            faas["trigger"]["type"] = "http"
            faas["trigger"]["request_id"] = self.event["requestContext"]["requestId"]
            service_context["origin"] = {"name": self.event["requestContext"]["domainName"]}
            service_context["origin"]["id"] = self.event["requestContext"]["apiId"]
            service_context["origin"]["version"] = self.event.get("version", "1.0")
            cloud_context["origin"] = {}
            cloud_context["origin"]["service"] = {"name": "api gateway"}
            if ".lambda-url." in self.event["requestContext"]["domainName"]:
                cloud_context["origin"]["service"]["name"] = "lambda url"
            cloud_context["origin"]["account"] = {"id": self.event["requestContext"]["accountId"]}
            cloud_context["origin"]["provider"] = "aws"
        elif self.source == "elb":
            elb_target_group_arn = self.event["requestContext"]["elb"]["targetGroupArn"]
            faas["trigger"]["type"] = "http"
            service_context["origin"] = {"name": elb_target_group_arn.split(":")[5].split("/")[1]}
            service_context["origin"]["id"] = elb_target_group_arn
            cloud_context["origin"] = {}
            cloud_context["origin"]["service"] = {"name": "elb"}
            cloud_context["origin"]["account"] = {"id": elb_target_group_arn.split(":")[4]}
            cloud_context["origin"]["region"] = elb_target_group_arn.split(":")[3]
            cloud_context["origin"]["provider"] = "aws"
        elif self.source == "sqs":
            record = self.event["Records"][0]
            faas["trigger"]["type"] = "pubsub"
            faas["trigger"]["request_id"] = record["messageId"]
            service_context["origin"] = {}
            service_context["origin"]["name"] = record["eventSourceARN"].split(":")[5]
            service_context["origin"]["id"] = record["eventSourceARN"]
            cloud_context["origin"] = {}
            cloud_context["origin"]["service"] = {"name": "sqs"}
            cloud_context["origin"]["region"] = record["awsRegion"]
            cloud_context["origin"]["account"] = {"id": record["eventSourceARN"].split(":")[4]}
            cloud_context["origin"]["provider"] = "aws"
            message_context["queue"] = {"name": service_context["origin"]["name"]}
            if "SentTimestamp" in record["attributes"]:
                message_context["age"] = {"ms": int((time.time() * 1000) - int(record["attributes"]["SentTimestamp"]))}
            if self.client.config.capture_body in ("request", "all") and "body" in record:
                message_context["body"] = record["body"]
            if self.client.config.capture_headers and record.get("messageAttributes"):
                headers = {}
                for k, v in record["messageAttributes"].items():
                    if v and v.get("stringValue"):
                        headers[k] = v.get("stringValue")
                if headers:
                    message_context["headers"] = headers
        elif self.source == "sns":
            record = self.event["Records"][0]
            faas["trigger"]["type"] = "pubsub"
            faas["trigger"]["request_id"] = record["Sns"]["TopicArn"]
            service_context["origin"] = {}
            service_context["origin"]["name"] = record["Sns"]["TopicArn"].split(":")[5]
            service_context["origin"]["id"] = record["Sns"]["TopicArn"]
            service_context["origin"]["version"] = record["EventVersion"]
            service_context["origin"]["service"] = {"name": "sns"}
            cloud_context["origin"] = {}
            cloud_context["origin"]["region"] = record["Sns"]["TopicArn"].split(":")[3]
            cloud_context["origin"]["account"] = {"id": record["Sns"]["TopicArn"].split(":")[4]}
            cloud_context["origin"]["provider"] = "aws"
            message_context["queue"] = {"name": service_context["origin"]["name"]}
            if "Timestamp" in record["Sns"]:
                message_context["age"] = {
                    "ms": int(
                        (
                            datetime.datetime.now()
                            - datetime.datetime.strptime(record["Sns"]["Timestamp"], r"%Y-%m-%dT%H:%M:%S.%fZ")
                        ).total_seconds()
                        * 1000
                    )
                }
            if self.client.config.capture_body in ("request", "all") and "Message" in record["Sns"]:
                message_context["body"] = record["Sns"]["Message"]
            if self.client.config.capture_headers and record["Sns"].get("MessageAttributes"):
                headers = {}
                for k, v in record["Sns"]["MessageAttributes"].items():
                    if v and v.get("Type") == "String":
                        headers[k] = v.get("Value")
                if headers:
                    message_context["headers"] = headers
        elif self.source == "s3":
            record = self.event["Records"][0]
            faas["trigger"]["type"] = "datasource"
            faas["trigger"]["request_id"] = record["responseElements"]["x-amz-request-id"]
            service_context["origin"] = {}
            service_context["origin"]["name"] = record["s3"]["bucket"]["name"]
            service_context["origin"]["id"] = record["s3"]["bucket"]["arn"]
            service_context["origin"]["version"] = record["eventVersion"]
            cloud_context["origin"] = {}
            cloud_context["origin"]["service"] = {"name": "s3"}
            cloud_context["origin"]["region"] = record["awsRegion"]
            cloud_context["origin"]["provider"] = "aws"

        metadata["service"] = {}
        # metadata["service"]["name"] = self.client.config.service_name
        metadata["service"]["framework"] = {"name": "AWS Lambda"}
        metadata["service"]["runtime"] = {
            "name": os.environ.get("AWS_EXECUTION_ENV"),
            "version": platform.python_version(),
        }
        metadata["service"]["version"] = self.client.config.service_version
        metadata["service"]["node"] = {"configured_name": os.environ.get("AWS_LAMBDA_LOG_STREAM_NAME")}
        # This is the one piece of metadata that requires deep merging. We add it manually
        # here to avoid having to deep merge in _transport.add_metadata()
        node = nested_key(self.client.get_service_info(), "node")
        if node:
            metadata["service"]["node"] = node

        metadata["cloud"] = {}
        metadata["cloud"]["provider"] = "aws"
        metadata["cloud"]["region"] = os.environ.get("AWS_REGION")
        metadata["cloud"]["service"] = {"name": "lambda"}
        metadata["cloud"]["account"] = {"id": arn.split(":")[4]}

        atatus.set_context(cloud_context, "cloud")
        atatus.set_context(service_context, "service")
        # faas doesn't actually belong in context, but we handle this in to_dict
        atatus.set_context(faas, "faas")
        if message_context:
            atatus.set_context(message_context, "message")
        self.client.add_extra_metadata(metadata)

    def send_partial_transaction(self) -> None:
        """
        We synchronously send the (partial) transaction to the Lambda Extension
        so that the transaction can be reported even if the lambda runtime times
        out before we can report the transaction.

        This is pretty specific to the HTTP transport. If we ever add other
        transports, we will need to clean this up.
        """
        global REGISTER_PARTIAL_TRANSACTIONS
        if (
            REGISTER_PARTIAL_TRANSACTIONS
            and os.environ.get("ATATUS_LAMBDA_APM_SERVER")
            and ("localhost" in self.client.config.server_url or "127.0.0.1" in self.client.config.server_url)
        ):
            transaction = execution_context.get_transaction()
            transport = self.client._transport
            logger.debug("Sending partial transaction and early metadata to the lambda extension...")
            data = transport._json_serializer({"metadata": self.client.build_metadata()}) + "\n"
            data += transport._json_serializer({"transaction": transaction.to_dict()})
            partial_transaction_url = urllib.parse.urljoin(
                self.client.config.server_url
                if self.client.config.server_url.endswith("/")
                else self.client.config.server_url + "/",
                "register/transaction",
            )
            try:
                transaction.pause_sampling = True
                transport.send(
                    data,
                    custom_url=partial_transaction_url,
                    custom_headers={
                        "x-elastic-aws-request-id": self.context.aws_request_id,
                        "Content-Type": "application/vnd.elastic.apm.transaction+ndjson",
                    },
                )
            except Exception as e:
                if re.match(r"HTTP [4,5]\d\d", str(e)):
                    REGISTER_PARTIAL_TRANSACTIONS = False
                    logger.info(
                        "APM Lambda Extension does not support partial transactions. "
                        "Disabling partial transaction registration."
                    )
                else:
                    logger.warning("Failed to send partial transaction to APM Lambda Extension", exc_info=True)
            finally:
                transaction.pause_sampling = False


def get_data_from_request(event: dict, capture_body: bool = False, capture_headers: bool = True) -> dict:
    """
    Capture context data from API gateway event
    """
    result = {}
    if capture_headers and "headers" in event:
        result["headers"] = event["headers"]

    method = (
        nested_key(event, "requestContext", "httpMethod")
        or nested_key(event, "requestContext", "http", "method")
        or nested_key(event, "httpMethod")
    )

    if not method:
        # Not API Gateway
        return result

    result["method"] = method
    if method in constants.HTTP_WITH_BODY and "body" in event:
        body = event["body"]
        if capture_body:
            if event.get("isBase64Encoded"):
                body = base64.b64decode(body)
            else:
                try:
                    jsonbody = json.loads(body)
                    body = jsonbody
                except Exception:
                    pass

        if body is not None:
            result["body"] = body if capture_body else "[REDACTED]"

    result["url"] = get_url_dict(event)
    return result


def get_data_from_response(response: dict, capture_headers: bool = True) -> dict:
    """
    Capture response data from lambda return
    """
    result = {}

    if "statusCode" in response:
        try:
            result["status_code"] = int(response["statusCode"])
        except ValueError:
            # statusCode wasn't formed as an int
            # we don't log here, as we will have already logged at transaction.result handling
            result["status_code"] = 500

    if capture_headers and "headers" in response:
        result["headers"] = response["headers"]
    return result


def get_url_dict(event: dict) -> dict:
    """
    Reconstruct URL from API Gateway
    """
    headers = event.get("headers") or {}
    protocol = headers.get("X-Forwarded-Proto", headers.get("x-forwarded-proto", "https"))
    host = headers.get("Host", headers.get("host", ""))
    stage = nested_key(event, "requestContext", "stage") or ""
    raw_path = event.get("rawPath", "")
    if stage:
        stage = "/" + stage
        raw_path = raw_path.split(stage)[-1]

    path = event.get("path", raw_path)
    port = headers.get("X-Forwarded-Port", headers.get("x-forwarded-port"))
    query = ""
    if "rawQueryString" in event:
        query = event["rawQueryString"]
    elif event.get("queryStringParameters"):
        if stage:  # api requires parameters encoding to build correct url
            query = "?" + urlencode(event["queryStringParameters"])
        else:  # for elb we do not have the stage
            query = "?" + "&".join(["{}={}".format(k, v) for k, v in event["queryStringParameters"].items()])

    url = protocol + "://" + host + stage + path + query

    url_dict = {
        "full": encoding.keyword_field(url),
        "protocol": protocol,
        "hostname": encoding.keyword_field(host),
        "pathname": encoding.keyword_field(stage + path),
    }

    if port:
        url_dict["port"] = port
    if query:
        url_dict["search"] = encoding.keyword_field(query)
    return url_dict



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import absolute_import

from atatus.instrumentation.packages.dbapi2 import (
    ConnectionProxy,
    CursorProxy,
    DbApi2Instrumentation,
    extract_signature,
)
from atatus.instrumentation.packages.psycopg2 import get_destination_info
from atatus.traces import capture_span


class PGCursorProxy(CursorProxy):
    provider_name = "postgresql"

    def _bake_sql(self, sql):
        # If this is a Composable object, use its `as_string` method.
        # See https://www.psycopg.org/psycopg3/docs/api/sql.html
        if hasattr(sql, "as_string"):
            sql = sql.as_string(self.__wrapped__)
        # If the sql string is already a byte string, we need to decode it using the connection encoding
        if isinstance(sql, bytes):
            sql = sql.decode(self.connection.info.encoding)
        return sql

    def extract_signature(self, sql):
        return extract_signature(sql)

    def __enter__(self):
        return PGCursorProxy(self.__wrapped__.__enter__(), destination_info=self._self_destination_info)

    @property
    def _self_database(self):
        return self.connection.info.dbname or ""


class PGConnectionProxy(ConnectionProxy):
    cursor_proxy = PGCursorProxy

    def __enter__(self):
        return PGConnectionProxy(self.__wrapped__.__enter__(), destination_info=self._self_destination_info)


class PsycopgInstrumentation(DbApi2Instrumentation):
    name = "psycopg"

    instrument_list = [("psycopg", "connect")]

    def call(self, module, method, wrapped, instance, args, kwargs):
        signature = "psycopg.connect"

        host, port = get_destination_info(kwargs.get("host"), kwargs.get("port"))
        database = kwargs.get("dbname")
        signature = f"{signature} {host}:{port}"
        destination_info = {
            "address": host,
            "port": port,
        }
        with capture_span(
            signature,
            span_type="db",
            span_subtype="postgresql",
            span_action="connect",
            leaf=True,
            extra={"destination": destination_info, "db": {"type": "sql", "instance": database}},
        ):
            return PGConnectionProxy(wrapped(*args, **kwargs), destination_info=destination_info)



import os
import sys
import subprocess


def _key_value_from_command(cmd, sep, successful_status=(0,)):
    d = {}
    for line in _command_by_line(cmd, successful_status=successful_status):
        l = [s.strip() for s in line.split(sep, 1)]
        if len(l) == 2:
            d[l[0]] = l[1]
    return d


def _command_by_line(cmd, successful_status=(0,)):
    ok, output = _getoutput(cmd, successful_status=successful_status)
    if not ok:
        return

    output = output.decode('ascii')

    for line in output.splitlines():
        yield line.strip()


def _getoutput(cmd, successful_status=(0,)):
    try:
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        output, _ = p.communicate()
        status = p.returncode
    except EnvironmentError as e:
        return False, ''
    if os.WIFEXITED(status) and os.WEXITSTATUS(status) in successful_status:
        return True, output
    return False, output


def darwin_hwinfo():
    hwinfo = {}
    hw = _key_value_from_command(['sysctl', 'hw'], sep=":")
    machdep_cpu = _key_value_from_command(['sysctl', 'machdep.cpu'], sep=":")
    cpu_count = 1
    cpu_freq = 0
    model_name = ""
    cpu_info = []
    mem_size = 0
    if hw:
        if 'hw.logicalcpu' in hw:
            cpu_count = int(hw['hw.logicalcpu'])
        if 'hw.cpufrequency' in hw:
            cpu_freq = hw['hw.cpufrequency']
        if 'hw.memsize' in hw:
            mem_size = int(hw['hw.memsize'])
    if machdep_cpu:
        if 'machdep.cpu.brand_string' in machdep_cpu:
            model_name = machdep_cpu['machdep.cpu.brand_string']
    for i in range(cpu_count):
        entry = {}
        entry["cores"] = 1
        entry["mhz"] = cpu_freq
        entry["model"] = model_name
        cpu_info.append(entry)

    hwinfo["cpu"] = cpu_info
    hwinfo["mem_size"] = mem_size
    return hwinfo


def linux_hwinfo():
    hwinfo = {}
    hwinfo["cpu"] = _linux_cpuinfo()
    hwinfo["mem_size"] = _linux_meminfo()
    bootid = _linux_bootid()
    if bootid:
        hwinfo["bootid"] = bootid
    return hwinfo


def _linux_bootid():
    try:
        fo = open('/proc/sys/kernel/random/boot_id')
    except EnvironmentError as e:
        return None
    else:
        bootid = fo.read().strip()
        fo.close()
        return bootid


def _linux_cpuinfo():
    try:
        fo = open('/proc/cpuinfo')
    except EnvironmentError as e:
        return [
            {
                "cores": "0",
                "mhz": 0,
                "model": "linux /proc/cpuinfo file not accessible"
            }
        ]
    else:
        info = [{}]
        for line in fo:
            name_value = [s.strip() for s in line.split(':', 1)]
            if len(name_value) != 2:
                continue
            name, value = name_value
            if not info or name in info[-1]:  # next processor
                info.append({})
            info[-1][name] = value
        fo.close()
        cpu_info = []
        for i in info:
            entry = {}
            entry["cores"] = 1
            if "cpu MHz" in i:
                entry["mhz"] = i["cpu MHz"]
            if "model name" in i:
                entry["model"] = i["model name"]
            cpu_info.append(entry)
        return cpu_info


def _linux_meminfo():
    try:
        fo = open('/proc/meminfo')
    except EnvironmentError as e:
        return -1
    else:
        info = [{}]
        for line in fo:
            name_value = [s.strip() for s in line.split(':', 1)]
            if len(name_value) != 2:
                continue
            name, value = name_value
            if not info or name in info[-1]:  # next processor
                info.append({})
            if value.endswith(' kB'):
                value = int(value[:-3])
                value = value * 1024  # to bytes
            info[-1][name] = value
        fo.close()
        mem_size = 0
        for i in info:
            if "MemTotal" in i:
                mem_size = i["MemTotal"]

        return mem_size


def default_hwinfo():
    hwinfo = {}
    hwinfo["cpu"] = [{
        "cores": "0",
        "mhz": 0,
        "model": sys.platform + " not supported"
    }]
    hwinfo["mem_size"] = 0
    return hwinfo



#  BSD 3-Clause License
#
#  Copyright (c) 2022, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import grpc

from atatus.conf.constants import OUTCOME

# see https://github.com/elastic/apm/blob/cf182bcfbe35b8586/specs/agents/tracing-instrumentation-grpc.md#outcome

STATUS_TO_OUTCOME = {
    grpc.StatusCode.OK: OUTCOME.SUCCESS,
    grpc.StatusCode.CANCELLED: OUTCOME.SUCCESS,  # Operation cancelled by client
    grpc.StatusCode.UNKNOWN: OUTCOME.FAILURE,  # Error of an unknown type, but still an error
    grpc.StatusCode.INVALID_ARGUMENT: OUTCOME.SUCCESS,  # Client-side error
    grpc.StatusCode.DEADLINE_EXCEEDED: OUTCOME.FAILURE,
    grpc.StatusCode.NOT_FOUND: OUTCOME.SUCCESS,  # Client-side error (similar to HTTP 404)
    grpc.StatusCode.ALREADY_EXISTS: OUTCOME.SUCCESS,  # Client-side error (similar to HTTP 409)
    grpc.StatusCode.PERMISSION_DENIED: OUTCOME.SUCCESS,  # Client authentication (similar to HTTP 403)
    grpc.StatusCode.RESOURCE_EXHAUSTED: OUTCOME.FAILURE,  # Likely used for server out of resources
    grpc.StatusCode.FAILED_PRECONDITION: OUTCOME.FAILURE,  # Similar to UNAVAILABLE
    grpc.StatusCode.ABORTED: OUTCOME.FAILURE,  # Similar to UNAVAILABLE
    grpc.StatusCode.OUT_OF_RANGE: OUTCOME.SUCCESS,  # Client-side error (similar to HTTP 416)
    grpc.StatusCode.UNIMPLEMENTED: OUTCOME.SUCCESS,  # Client called a non-implemented feature
    grpc.StatusCode.INTERNAL: OUTCOME.FAILURE,  # Internal error (similar to HTTP 500)
    grpc.StatusCode.UNAVAILABLE: OUTCOME.FAILURE,  # Transient error, client may retry with backoff
    grpc.StatusCode.DATA_LOSS: OUTCOME.FAILURE,  # Lost data should always be reported
    grpc.StatusCode.UNAUTHENTICATED: OUTCOME.SUCCESS,  # Client-side authentication (similar to HTTP 401)
}



import time
import json
import sched
import random
import threading
import traceback
from atatus.utils.logging import get_logger
from atatus.utils import encoding
from .transport import BaseTransport
from .layer import Layer
from .hist import TxnHist

import urllib.parse
from atatus.utils.module_import import import_string
from atatus.conf.constants import ERROR, SPAN, TRANSACTION, METRICSET

import atatus
from collections import namedtuple
SpanTiming = namedtuple('SpanTiming', 'start end')
ERROR_LIMIT = 200

class Txn(Layer):
    def __init__(self, type, kind, duration, background):
        super(Txn, self).__init__(type, kind, duration)
        self.spans = {}
        self.background = background


class Collector(object):
    def __init__(self, client, transport_kwargs):

        metadata = None
        if "metadata" in transport_kwargs:
            metadata = transport_kwargs["metadata"]

        analytics_processors = None
        if "analytics_processors" in transport_kwargs:
            analytics_processors = transport_kwargs["analytics_processors"]

        self._client = client
        self._config = self._client.config
        self._config.llm_active = False
        cls = self.__class__
        self._error_logger = get_logger("atatus.errors")
        self._logger = get_logger("%s.%s" % (cls.__module__, cls.__name__))

        if not self._config.app_name and not self._config.license_key:
            self._error_logger.error("Error: Atatus configuration app_name and license_key are missing!")
        elif not self._config.license_key:
            self._error_logger.error("Error: Atatus configuration license_key is missing!")
        elif not self._config.app_name:
            self._error_logger.error("Error: Atatus configuration app_name is missing!")

        self._spans = {}
        self._txns_lock = threading.Lock()
        self._txns_agg = {}
        self._txn_hist_agg = {}
        self._traces_agg = []
        self._error_metrics_agg = {}
        self._error_requests_agg = []

        self._errors_lock = threading.Lock()
        self._errors_agg = []

        self._metrics_lock = threading.Lock()
        self._metrics_agg = []

        self._analytics = self._config.analytics
        self._analytics_lock = threading.Lock()
        self._analytics_agg = []

        self._hostinfo_response = {}
        self._hostinfo_response["analytics"] = True

        self._transport = BaseTransport(self._config, metadata)
        self._collect_counter = 0

        self._scheduler = None
        self._start_time = int(time.time() * 1000)
        self._worker = None
        self._closed = False
        # self._pid = None
        self._thread_starter_lock = threading.Lock()
        self._analytics_processors = analytics_processors if analytics_processors is not None else []

        self._tracing = self._config.tracing
        self._error_limit = self._config.errorLimit if self._config.errorLimit else ERROR_LIMIT
        self._hostinfo_response["tracing"] = True
        self._transaction_ignore_urls = self._config.transaction_ignore_urls.copy()

        if self._hostinfo_response["tracing"] == True and self._tracing is True:
            dt_endpoint_url = urllib.parse.urljoin(
                self._config.notify_host,
                "/track/traces/spans",
            )        
            transport_class = import_string(self._config.transport_class)
            self._dt_transport = transport_class(url=dt_endpoint_url, client=self._client, **transport_kwargs)
            self._config.transport = self._dt_transport
            self._client._thread_managers["transport"] = self._dt_transport 

        self._ensure_collector_running()      


    def _start_collector_thread(self):
        if (not self._worker or not self._worker.is_alive()) and not self._closed:
            try:
                self._worker = threading.Thread(target=self._worker_thread, name="Atatus Collector Worker")
                self._worker.setDaemon(True)
                self._worker_exit = threading.Event()
                self._worker.start()
            except RuntimeError:
                pass

    def _ensure_collector_running(self):
        with self._thread_starter_lock:
            self._start_collector_thread()
            # check if self._pid is the same as os.getpid(). If they are not the same, it means that our
            # process was forked at some point, so we need to start another processor thread
            # if self._pid != os.getpid():
            #     self._start_collector_thread()
            #     self._pid = os.getpid()

    def close(self):
        if self._closed:
            return
        if not self._worker or not self._worker.is_alive():
            return
        self._closed = True
        self._worker_exit.set()
        if self._worker.is_alive():
            self._worker.join(10)

    def _worker_timefunc(self):
        if self._worker_exit.isSet():
            return float("inf")
        return time.time()

    def _worker_thread(self):
        self._scheduler = sched.scheduler(self._worker_timefunc, self._worker_exit.wait)
        self._start_time = int(time.time() * 1000)
        self._collect()
        try:
            self._scheduler.run()
        except Exception as e:
            self._error_logger.error("Atatus worker failed with exception: %r", e)
            self._error_logger.error(traceback.format_exc())

    def _collect(self):
        start_time = self._start_time

        if not self._worker_exit.isSet():
            self._start_time = int(time.time() * 1000)
            self._scheduler.enter(60.0, 1, self._collect, ())

        end_time = int(time.time() * 1000)

        if not self._config.app_name or not self._config.license_key:
            if not self._config.app_name and not self._config.license_key:
                self._error_logger.error("Error: Atatus configuration app_name and license_key are missing!")
            elif not self._config.license_key:
                self._error_logger.error("Error: Atatus configuration license_key is missing!")
            elif not self._config.app_name:
                self._error_logger.error("Error: Atatus configuration app_name is missing!")
            return

        with self._txns_lock:
            txns_data = self._txns_agg
            self._txns_agg = {}

            txn_hist_data = self._txn_hist_agg
            self._txn_hist_agg = {}

            traces_data = self._traces_agg
            self._traces_agg = []

            error_metrics_data = self._error_metrics_agg
            self._error_metrics_agg = {}

            error_requests_data = self._error_requests_agg
            self._error_requests_agg = []

        with self._errors_lock:
            errors_data = self._errors_agg
            self._errors_agg = []

        with self._metrics_lock:
            metrics_data = self._metrics_agg
            self._metrics_agg = []

        with self._analytics_lock:
            analytics_data = self._analytics_agg
            self._analytics_agg = []

        try:
            if self._collect_counter % 30 == 0:
                self._hostinfo_response = self._transport.hostinfo(start_time)

                self._config.transaction_ignore_urls = self._transaction_ignore_urls + self._hostinfo_response["ignoreTxnUrlPatterns"]
                
                self._collect_counter = 0
            self._collect_counter += 1
            
            if txns_data:
                self._transport.txns(start_time, end_time, txns_data)

            if txn_hist_data:
                self._transport.txn_hist(start_time, end_time, txn_hist_data)

            if traces_data:
                for trace in traces_data:
                    individual_trace_data = [trace]
                    self._transport.traces(start_time, end_time, individual_trace_data)

            if error_metrics_data:
                self._transport.error_metrics(start_time, end_time, error_metrics_data, error_requests_data)

            if errors_data:
                self._transport.errors(start_time, end_time, errors_data)

            if metrics_data:
                self._transport.metrics(start_time, end_time, metrics_data)

            if analytics_data:
                if "analytics" in self._hostinfo_response:
                    if self._hostinfo_response["analytics"] == True and self._analytics == True:
                        analytics_sanitized_data = []
                        analytics_payload_len = 0
                        for event in analytics_data:
                            sevent = self._process_analytics_event(event)
                            if sevent != None:
                                analytics_sanitized_data.append(sevent)
                                if "body_len" in sevent:
                                    analytics_payload_len += sevent["body_len"]
                                if analytics_payload_len > 6 * 1024 * 1024:
                                    self._transport.analytics(start_time, end_time, analytics_sanitized_data)
                                    analytics_sanitized_data = []
                                    analytics_payload_len = 0
                        if len(analytics_sanitized_data) > 0:
                            self._transport.analytics(start_time, end_time, analytics_sanitized_data)

        except Exception as e:
            self._error_logger.debug("Atatus collect failed with exception: %r" % e)
            self._error_logger.debug(traceback.format_exc())

    def _process_analytics_event(self, data):
        # Run the data through processors
        for processor in self._analytics_processors:
            try:
                data = processor(self._client, data)
                if not data:
                    return None
            except Exception:
                return None
        return data

    def add_error(self, error):
        self._ensure_collector_running()

        skip_event = False
        if self._config.skip_processor is not None:
            error_txn_name = ""
            if "event_transaction_name" in error:
                error_txn_name = error["event_transaction_name"]

            error_url = ""
            if "context" in error and \
                "request" in error["context"] and \
                "url" in error["context"]["request"] and \
                "full" in error["context"]["request"]["url"]:
                    error_url = error["context"]["request"]["url"]["full"]
            try:
                skip_event_dict = {}
                skip_event_dict["name"] = error_txn_name
                skip_event_dict["url"] = error_url
                skip_event = self._config.skip_processor(skip_event_dict)
            except Exception as e:
                self._error_logger.debug("error skip function failed with exception: %r" % e)
                pass

        if skip_event == True:
            if "event_transaction_name" in error:
                self._error_logger.debug("skipping error event: %r" % error["event_transaction_name"])
            return

        with self._errors_lock:
            limit = self._hostinfo_response["errorLimit"] if "errorLimit" in self._hostinfo_response else self._error_limit
            if len(self._errors_agg) < limit:
                self._errors_agg.append(error)
            else:
                self._errors_agg = self._errors_agg[:limit]
                i = random.randrange(limit)
                self._errors_agg[i] = error

    def add_metricset(self, metricset):
        self._ensure_collector_running()

        if "samples" not in metricset:
            return
        s = metricset["samples"]

        if not all(k in s for k in ("system.cpu.total.norm.pct", "system.memory.actual.free", "system.memory.total", "system.process.cpu.total.norm.pct", "system.process.memory.size", "system.process.memory.rss.bytes")):
            return

        if not all("value" in s[k] for k in ("system.cpu.total.norm.pct", "system.memory.actual.free", "system.memory.total", "system.process.cpu.total.norm.pct", "system.process.memory.size", "system.process.memory.rss.bytes")):
            return

        metric = {
            "system.cpu.total.norm.pct": s["system.cpu.total.norm.pct"]["value"],
            "system.memory.actual.free": s["system.memory.actual.free"]["value"],
            "system.memory.total": s["system.memory.total"]["value"],
            "system.process.cpu.total.norm.pct": s["system.process.cpu.total.norm.pct"]["value"],
            "system.process.memory.size": s["system.process.memory.size"]["value"],
            "system.process.memory.rss.bytes": s["system.process.memory.rss.bytes"]["value"]
        }

        with self._metrics_lock:
            self._metrics_agg.append(metric)

    def add_analytics(self, txn):
        analytics_txn = {}

        if not all(k in txn for k in ("name", "id", "timestamp", "duration")):
            return

        analytics_txn["timestamp"] = (txn["timestamp"] // 1000)
        analytics_txn["txnId"] = txn["id"]
        if "trace_id" in txn:
            analytics_txn["traceId"] = txn["trace_id"]
        
        analytics_txn["name"] = txn["name"]
        analytics_txn["duration"] = txn["duration"]

        requestBodyLen = 0
        responseBodyLen = 0
        if "context" in txn and \
            "request" in txn["context"]:

            if "method" in txn["context"]["request"]:
                analytics_txn["method"] = txn["context"]["request"]["method"]

            if "headers" in txn["context"]["request"]:
                analytics_txn["requestHeaders"] = txn["context"]["request"]["headers"]
                if "User-Agent" in analytics_txn["requestHeaders"]:
                    analytics_txn["userAgent"] = analytics_txn["requestHeaders"]["User-Agent"]
                elif "user-agent" in analytics_txn["requestHeaders"]:
                    analytics_txn["userAgent"] = analytics_txn["requestHeaders"]["user-agent"]

            if "body" in txn["context"]["request"]:
                if txn["context"]["request"]["body"] != "[REDACTED]":
                    if txn["context"]["request"]["body"] != "":
                        instance_dict_convert = False
                        string_body = ""
                        if isinstance(txn["context"]["request"]["body"], dict):
                            try:
                                string_body = json.dumps(txn["context"]["request"]["body"])
                                instance_dict_convert = True
                            except Exception:
                                pass
                        if instance_dict_convert == False:
                            string_body = str(txn["context"]["request"]["body"])

                        analytics_txn["requestBody"] = encoding.body_field(string_body)
                        requestBodyLen = len(analytics_txn["requestBody"])

            if "url" in txn["context"]["request"]:
                if "full" in txn["context"]["request"]["url"]:
                    analytics_txn["url"] = txn["context"]["request"]["url"]["full"]

            if "socket" in txn["context"]["request"]:
                if "remote_address" in txn["context"]["request"]["socket"]:
                    analytics_txn["ip"] = txn["context"]["request"]["socket"]["remote_address"]

        if "context" in txn and \
            "response" in txn["context"]:
                if "body" in txn["context"]["response"]:
                    analytics_txn["responseBody"] = txn["context"]["response"]["body"]
                    responseBodyLen = len(analytics_txn["responseBody"])

        if "context" in txn and \
            "custom_response" in txn["context"]:
                if "body" in txn["context"]["custom_response"]:
                    analytics_txn["responseBody"] = txn["context"]["custom_response"]["body"]
                    responseBodyLen = len(analytics_txn["responseBody"])

        if "context" in txn and \
            "response" in txn["context"]:

            if "headers" in txn["context"]["response"]:
                analytics_txn["responseHeaders"] = txn["context"]["response"]["headers"]

            if "status_code" in txn["context"]["response"]:
                analytics_txn["statusCode"] = txn["context"]["response"]["status_code"]

            # analytics_txn["customData"] = txn.context.custom

        if "context" in txn and \
            "custom" in txn["context"] and \
            isinstance(txn["context"]["custom"], dict):
            custom_data = {}
            for k, v in txn["context"]["custom"].items():
                try:
                    custom_data[str(k)] = str(v)
                except Exception as e:
                    continue
            analytics_txn["customData"] = custom_data

        if "context" in txn and \
            "user" in txn["context"]:
            
            if "id" in txn["context"]["user"]:
                analytics_txn["userId"] = str(txn["context"]["user"]["id"])

            if "username" in txn["context"]["user"]:
                analytics_txn["userName"] = txn["context"]["user"]["username"]

            if "email" in txn["context"]["user"]:
                analytics_txn["userEmail"] = txn["context"]["user"]["email"]

        if "context" in txn and \
            "company" in txn["context"]:

            if "id" in txn["context"]["company"]:
                analytics_txn["companyId"] = str(txn["context"]["company"]["id"])

        analytics_txn["body_len"] = requestBodyLen + responseBodyLen
        analytics_txn["direction"] = "incoming"
        with self._analytics_lock:
            if len(self._analytics_agg) < 10000:
                skip_event = False

                if self._config.analytics_skip_processor is not None:
                    try:
                        skip_event = self._config.analytics_skip_processor(analytics_txn)
                    except Exception as e:
                        self._error_logger.debug("analytics skip function failed with exception: %r" % e)
                        pass

                if skip_event == False:
                    if self._config.analytics_mask_processor is not None:
                        try:
                            self._config.analytics_mask_processor(analytics_txn)
                        except Exception as e:
                            self._error_logger.debug("mask function failed with exception: %r" % e)
                            pass
                    self._analytics_agg.append(analytics_txn)
                else:
                    if "name" in analytics_txn:
                        self._error_logger.debug("skipping analytics event: %r" % analytics_txn["name"])

    def add_analytics_outgoing(self, txn, span):
        analytics_txn = {}

        if not all(k in txn for k in ("name", "id", "timestamp", "duration")):
            return

        if not all(k in span for k in ("name", "timestamp", "duration", "context")):
            return

        if "analytics_captured" not in span["context"]:
            return

        analytics_txn["timestamp"] = (txn["timestamp"] // 1000)
        analytics_txn["txnId"] = txn["id"]
        if "trace_id" in txn:
            analytics_txn["traceId"] = txn["trace_id"]
        
        analytics_txn["name"] = span["name"]
        analytics_txn["duration"] = span["duration"]

        requestBodyLen = 0
        responseBodyLen = 0
        if "context" in span and \
            "http" in span["context"] and \
            "request" in span["context"]["http"]:

            if "method" in span["context"]["http"]["request"]:
                analytics_txn["method"] = span["context"]["http"]["request"]["method"]

            if "headers" in span["context"]["http"]["request"]:
                analytics_txn["requestHeaders"] = span["context"]["http"]["request"]["headers"]
                if "User-Agent" in analytics_txn["requestHeaders"]:
                    analytics_txn["userAgent"] = analytics_txn["requestHeaders"]["User-Agent"]
                elif "user-agent" in analytics_txn["requestHeaders"]:
                    analytics_txn["userAgent"] = analytics_txn["requestHeaders"]["user-agent"]

            if "body" in span["context"]["http"]["request"]:
                if span["context"]["http"]["request"]["body"] != "[REDACTED]":
                    if span["context"]["http"]["request"]["body"] != "":
                        instance_dict_convert = False
                        string_body = ""
                        if isinstance(span["context"]["http"]["request"]["body"], dict):
                            try:
                                string_body = json.dumps(span["context"]["http"]["request"]["body"])
                                instance_dict_convert = True
                            except Exception:
                                pass
                        if instance_dict_convert == False:
                            string_body = str(span["context"]["http"]["request"]["body"])

                        analytics_txn["requestBody"] = encoding.body_field(string_body)
                        requestBodyLen = len(analytics_txn["requestBody"])

        if "context" in span and \
            "http" in span["context"] and \
            "response" in span["context"]["http"]:
                if "headers" in span["context"]["http"]["response"]:
                     analytics_txn["responseHeaders"] = span["context"]["http"]["response"]["headers"]

                if "body" in span["context"]["http"]["response"]:
                    analytics_txn["responseBody"] = span["context"]["http"]["response"]["body"]
                    responseBodyLen = len(analytics_txn["responseBody"])

        if "context" in span and \
            "http" in span["context"]:
            
            if "url" in span["context"]["http"]:
                analytics_txn["url"] = span["context"]["http"]["url"]
            if "status_code" in span["context"]["http"]:
                analytics_txn["statusCode"] = span["context"]["http"]["status_code"]

        if "context" in txn and \
            "user" in txn["context"]:
            
            if "id" in txn["context"]["user"]:
                analytics_txn["userId"] = str(txn["context"]["user"]["id"])

            if "username" in txn["context"]["user"]:
                analytics_txn["userName"] = txn["context"]["user"]["username"]

            if "email" in txn["context"]["user"]:
                analytics_txn["userEmail"] = txn["context"]["user"]["email"]

        if "context" in txn and \
            "company" in txn["context"]:

            if "id" in txn["context"]["company"]:
                analytics_txn["companyId"] = str(txn["context"]["company"]["id"])

        analytics_txn["body_len"] = requestBodyLen + responseBodyLen
        analytics_txn["direction"] = "outgoing"
        with self._analytics_lock:
             if len(self._analytics_agg) < 10000:
                skip_event = False
                if self._config.analytics_skip_processor is not None:
                    try:
                        skip_event = self._config.analytics_skip_processor(analytics_txn)
                    except Exception as e:
                        self._error_logger.debug("analytics outgoing skip function failed with exception: %r" % e)
                        pass

                if skip_event == False:
                    if self._config.analytics_mask_processor is not None:
                        try:
                            self._config.analytics_mask_processor(analytics_txn)
                        except Exception as e:
                            self._error_logger.debug("mask function failed with exception: %r" % e)
                            pass
                    self._analytics_agg.append(analytics_txn)
                else:
                    if "name" in analytics_txn:
                        self._error_logger.debug("skipping analytics event: %r" % analytics_txn["name"])

    def add_span(self, span):

        if "tracing" in self._hostinfo_response and self._hostinfo_response["tracing"] == True and self._tracing is True:
            if not self._config.llm_active and span['type'] == 'llmobs':
                self._config.llm_active = True
            self._dt_transport.queue(SPAN, span)

        if not all(k in span for k in ("transaction_id", "name", "type", "subtype", "duration")):
            return

        span_id = span["transaction_id"]
        if span_id not in self._spans:
            self._spans[span_id] = [span]
        else:
            self._spans[span_id].append(span)

    def add_txn(self, txn):
        is_tracing = "tracing" in self._hostinfo_response and self._hostinfo_response["tracing"] == True and self._tracing is True
        if is_tracing:
            if not self._config.llm_active and txn['type'] == 'llmobs':
                self._config.llm_active = True
            self._dt_transport.queue(TRANSACTION, txn)

        self._ensure_collector_running()

        if not all(k in txn for k in ("name", "id", "timestamp", "duration")):
            return
        txn_name = txn["name"]
        if not txn_name:
            txn_id = txn["id"]
            if txn_id:
                if txn_id in self._spans:
                    del self._spans[txn_id]
            return

        if txn["duration"] <= 0:
            return

        skip_event = False
        if self._config.skip_processor is not None:
            txn_url = ""
            if "context" in txn and \
                "request" in txn["context"] and \
                "url" in txn["context"]["request"] and \
                "full" in txn["context"]["request"]["url"]:
                    txn_url = txn["context"]["request"]["url"]["full"]
            try:
                skip_event_dict = {}
                skip_event_dict["name"] = txn_name
                skip_event_dict["url"] = txn_url
                skip_event = self._config.skip_processor(skip_event_dict)
            except Exception as e:
                self._error_logger.debug("txn skip function failed with exception: %r" % e)
                pass

        if skip_event == True:
            self._error_logger.debug("skipping txn event: %r" % txn_name)
            txn_id = txn["id"]
            if txn_id in self._spans:
                del self._spans[txn_id]
            return

        with self._txns_lock:
            if self._config.framework_name:
                txn_type = self._config.framework_name
            else:
                txn_type = atatus.PYTHON_AGENT

            background = False
            if "type" in txn:
                if txn["type"] == 'celery':
                    background=True

            if txn_name not in self._txns_agg:
                self._txns_agg[txn_name] = Txn(txn_type, atatus.PYTHON_AGENT, txn["duration"], background=background)
            else:
                self._txns_agg[txn_name].aggregate(txn["duration"])

            if background is False and txn["duration"] <= 150*1000.0:
                if txn_name not in self._txn_hist_agg:
                    self._txn_hist_agg[txn_name] = TxnHist(txn_type, atatus.PYTHON_AGENT, txn["duration"])
                else:
                    self._txn_hist_agg[txn_name].aggregate(txn["duration"])

            spans_present = False

            txn_id = txn["id"]
            python_time = 0
            spans_tuple = []
            if txn_id in self._spans:
                spans_present = True
                for span in self._spans[txn_id]:
                    if not all(k in span for k in ("name", "type", "subtype", "timestamp", "duration")):
                        continue
                    span_name = span["name"]
                    if not span_name:
                        continue
                    if span["timestamp"] >= txn["timestamp"]:
                        timestamp = ((span["timestamp"] - txn["timestamp"]) / 1000)
                        spans_tuple.append(SpanTiming(timestamp, timestamp + span["duration"]))

                        if self._config.analytics_capture_outgoing is True:
                            if span["type"] == "external" and span["subtype"] == "http":
                                if "analytics" in self._hostinfo_response:
                                    if self._hostinfo_response["analytics"] == True and self._analytics == True:
                                        if background == False:
                                            self.add_analytics_outgoing(txn, span)

                        if span_name not in self._txns_agg[txn_name].spans:
                            kind = Layer.kinds_dict.get(span["type"], span["type"])
                            type = Layer.types_dict.get(span["subtype"], span["subtype"])
                            self._txns_agg[txn_name].spans[span_name] = Layer(type, kind, span["duration"])
                        else:
                            self._txns_agg[txn_name].spans[span_name].aggregate(span["duration"])

            if len(spans_tuple) == 0:
                python_time = txn["duration"]
            else:
                spans_tuple.sort(key=lambda x: x.start)
                python_time = spans_tuple[0].start
                span_end = spans_tuple[0].end
                j = 0
                while j < len(spans_tuple):
                    if spans_tuple[j].start > span_end:
                        python_time += spans_tuple[j].start - span_end
                        span_end = spans_tuple[j].end
                    else:
                        if spans_tuple[j].end > span_end:
                            span_end = spans_tuple[j].end
                    j += 1
                if txn["duration"] > span_end:
                    python_time += txn["duration"] - span_end

            if python_time > 0:
                self._txns_agg[txn_name].spans[atatus.PYTHON_AGENT] = Layer(
                    atatus.PYTHON_AGENT, atatus.PYTHON_AGENT, python_time)

            if spans_present is True or python_time > 0:
                if txn["duration"] >= (self._config.trace_threshold.total_seconds() * 1000):
                    trace_txn = txn.copy()
                    if spans_present is True:
                        trace_txn["spans"] = self._spans[txn_id]
                    if python_time > 0:
                        trace_txn["python_time"] = python_time
                    trace_txn["background"] = background

                    if len(self._traces_agg) < 5:
                        self._traces_agg.append(trace_txn)
                    else:
                        i = random.randrange(5)
                        self._traces_agg[i] = trace_txn

            if spans_present:
                del self._spans[txn_id]

            if "context" in txn and \
               "response" in txn["context"] and \
               "status_code" in txn["context"]["response"]:

                status_code = txn["context"]["response"]["status_code"]

                if status_code >= 400 and (is_tracing or status_code != 404):
                    if txn_name not in self._error_metrics_agg:
                        self._error_metrics_agg[txn_name] = {status_code: 1}
                    else:
                        if status_code not in self._error_metrics_agg[txn_name]:
                            self._error_metrics_agg[txn_name][status_code] = 1
                        else:
                            self._error_metrics_agg[txn_name][status_code] += 1

                    if len(self._error_requests_agg) < ERROR_LIMIT:
                        self._error_requests_agg.append({"name": txn_name, "timestamp": txn["timestamp"], "context": txn["context"]})
                    else:
                        i = random.randrange(ERROR_LIMIT)
                        self._error_requests_agg[i] = {"name": txn_name, "timestamp": txn["timestamp"], "context": txn["context"]}

            if "analytics" in self._hostinfo_response:
                if self._hostinfo_response["analytics"] == True and self._analytics == True:
                    if background == False:
                        self.add_analytics(txn)



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import atatus
from atatus.conf import constants

try:
    import tornado
except ImportError:
    pass


def get_data_from_request(request_handler, request, config, event_type):
    """
    Capture relevant data from a tornado.httputil.HTTPServerRequest
    """
    result = {
        "method": request.method,
        "socket": {"remote_address": request.remote_ip},
        "cookies": request.cookies,
        "http_version": request.version,
    }
    if config.capture_headers:
        result["headers"] = dict(request.headers)
    if request.method in constants.HTTP_WITH_BODY:
        if tornado.web._has_stream_request_body(request_handler.__class__):
            # Body is a future and streaming is expected to be handled by
            # the user in the RequestHandler.data_received function.
            # Currently not sure of a safe way to get the body in this case.
            result["body"] = "[STREAMING]" if config.capture_body in ("all", "request") else "[REDACTED]"
        else:
            body = None
            try:
                body = tornado.escape.json_decode(request.body)
            except Exception:
                body = str(request.body, errors="ignore")

            if body is not None:
                result["body"] = body if config.capture_body in ("all", "request") else "[REDACTED]"

    result["url"] = atatus.utils.get_url_dict(request.full_url())
    return result


def get_data_from_response(request_handler, config, event_type):
    result = {}

    result["status_code"] = request_handler.get_status()

    if config.capture_headers and request_handler._headers:
        headers = request_handler._headers
        result["headers"] = {key: ";".join(headers.get_list(key)) for key in headers.keys()}
        pass
    return result



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from opentelemetry.trace import SpanKind
from opentelemetry.trace.span import SpanContext

import atatus.conf.constants as constants
from atatus.utils.disttracing import TraceParent, TracingOptions


def get_traceparent(otel_spancontext: SpanContext) -> TraceParent:
    """
    Create an atatus TraceParent object from an opentelemetry SpanContext
    """
    trace_id = None
    span_id = None
    is_sampled = False
    if otel_spancontext and otel_spancontext.is_valid:
        trace_id = otel_spancontext.trace_id
        span_id = otel_spancontext.span_id
        is_sampled = otel_spancontext.trace_flags.sampled
        tracestate = otel_spancontext.trace_state
    if trace_id:
        traceparent = TraceParent(
            constants.TRACE_CONTEXT_VERSION,
            "%032x" % trace_id,
            "%016x" % span_id,
            TracingOptions(recorded=is_sampled),
            tracestate=tracestate.to_header(),
        )
        return traceparent
    else:
        return None


def get_span_kind(kind: SpanKind) -> str:
    """
    Converts a SpanKind to the string representation
    """
    if kind == SpanKind.CLIENT:
        return "CLIENT"
    elif kind == SpanKind.CONSUMER:
        return "CONSUMER"
    elif kind == SpanKind.INTERNAL:
        return "INTERNAL"
    elif kind == SpanKind.PRODUCER:
        return "PRODUCER"
    elif kind == SpanKind.SERVER:
        return "SERVER"
    else:
        return "INTERNAL"



import math

class TxnHist(object):
    def __init__(self, type, kind, duration):
        self.type = type
        self.kind = kind
        self.hist = {}
        self.hist[self._find_bin(duration)] = 1

    def aggregate(self, duration):
        hbin = self._find_bin(duration)
        if hbin not in self.hist:
            self.hist[hbin] = 1
        else:
            self.hist[hbin] += 1

    def _round_bin(self, duration, unit):
        return math.floor(duration/float(unit)) * unit

    def _find_bin(self, duration):
        seconds = 1000
        hbin = 0

        if duration < 0:
            hbin = 0
        elif duration < 2*seconds:
            hbin = self._round_bin(duration, 10)
        elif duration < 10*seconds:
            hbin = self._round_bin(duration, 25)
        elif duration < 20*seconds:
            hbin = self._round_bin(duration, 50)
        elif duration < 50*seconds:
            hbin = self._round_bin(duration, 100)
        elif duration < 80*seconds:
            hbin = self._round_bin(duration, 200)
        elif duration < 100*seconds:
            hbin = self._round_bin(duration, 250)
        else:
            hbin = self._round_bin(duration, 1000)

        return hbin


#  BSD 3-Clause License
#
#  Copyright (c) 2012, the Sentry Team, see AUTHORS for more details
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE

import base64
import os
import re
import socket
import urllib.parse
from functools import partial, partialmethod
from types import FunctionType
from typing import Pattern

from atatus.conf import constants
from atatus.utils import encoding

partial_types = (partial, partialmethod)


default_ports = {"https": 443, "http": 80, "postgresql": 5432, "mysql": 3306, "mssql": 1433}
fqdn = None


def varmap(func, var, context=None, name=None, **kwargs):
    """
    Executes ``func(key_name, value)`` on all values,
    recursively discovering dict and list scoped
    values.
    """
    if context is None:
        context = set()
    objid = id(var)
    if objid in context:
        return func(name, "<...>", **kwargs)
    context.add(objid)

    # Apply func() before recursion, so that `shorten()` doesn't have to iterate over all the trimmed values
    ret = func(name, var, **kwargs)
    if isinstance(ret, dict):
        # iterate over a copy of the dictionary to avoid "dictionary changed size during iteration" issues
        ret = dict((k, varmap(func, v, context, k, **kwargs)) for k, v in ret.copy().items())
    elif isinstance(ret, (list, tuple)):
        # Apply func() before recursion, so that `shorten()` doesn't have to iterate over all the trimmed values
        ret = [varmap(func, f, context, name, **kwargs) for f in ret]
    context.remove(objid)
    return ret


def get_name_from_func(func: FunctionType) -> str:
    # partials don't have `__module__` or `__name__`, so we use the values from the "inner" function
    if isinstance(func, partial_types):
        return "partial({})".format(get_name_from_func(func.func))
    elif hasattr(func, "_partialmethod") and hasattr(func._partialmethod, "func"):
        return "partial({})".format(get_name_from_func(func._partialmethod.func))

    module = func.__module__

    if hasattr(func, "view_class"):
        view_name = func.view_class.__name__
    elif hasattr(func, "__name__"):
        view_name = func.__name__
    else:  # Fall back if there's no __name__
        view_name = func.__class__.__name__

    return "{0}.{1}".format(module, view_name)


def build_name_with_http_method_prefix(name, request):
    return " ".join((request.method, name)) if name else name


def is_master_process() -> bool:
    # currently only recognizes uwsgi master process
    try:
        import uwsgi

        return os.getpid() == uwsgi.masterpid()
    except ImportError:
        return False


def get_url_dict(url: str) -> dict:
    parse_result = urllib.parse.urlparse(url)

    url_dict = {
        "full": encoding.keyword_field(url),
        "protocol": parse_result.scheme + ":",
        "hostname": encoding.keyword_field(parse_result.hostname),
        "pathname": encoding.keyword_field(parse_result.path),
    }

    port = None if parse_result.port is None else str(parse_result.port)

    if port:
        url_dict["port"] = port
    if parse_result.query:
        url_dict["search"] = encoding.keyword_field("?" + parse_result.query)
    return url_dict


def sanitize_url(url: str) -> str:
    if "@" not in url:
        return url
    parts = urllib.parse.urlparse(url)
    return url.replace("%s:%s" % (parts.username, parts.password), "%s:%s" % (parts.username, constants.MASK_URL))


def get_host_from_url(url: str) -> str:
    parsed_url = urllib.parse.urlparse(url)
    host = parsed_url.hostname or " "

    if parsed_url.port and default_ports.get(parsed_url.scheme) != parsed_url.port:
        host += ":" + str(parsed_url.port)

    return host

def get_external_headers(headers):
    try:
        headers = {k: v for k, v in headers.items()}
    except:
        headers = {}
    return headers

def external_body_decode(body):
    if isinstance(body, bytes):
        try:
            string_data = body.decode('UTF-8')
            return string_data
        except:
            pass
    return body

def url_to_destination_resource(url: str) -> str:
    try:
        parts = urllib.parse.urlsplit(url)
    except ValueError:
        return url
    hostname = parts.hostname if parts.hostname else ""
    # preserve brackets for IPv6 URLs
    if "://[" in url:
        hostname = "[%s]" % hostname
    try:
        port = parts.port
    except ValueError:
        # Malformed port, just use None rather than raising an exception
        port = None
    default_port = default_ports.get(parts.scheme, None)
    name = "%s://%s" % (parts.scheme, hostname)
    resource = hostname
    if not port and parts.scheme in default_ports:
        port = default_ports[parts.scheme]
    if port:
        if port != default_port:
            name += ":%d" % port
        resource += ":%d" % port
    return resource


def read_pem_file(file_obj) -> bytes:
    cert = b""
    for line in file_obj:
        if line.startswith(b"-----BEGIN CERTIFICATE-----"):
            break
    # scan until we find the first END CERTIFICATE marker
    for line in file_obj:
        if line.startswith(b"-----END CERTIFICATE-----"):
            break
        cert += line.strip()
    return base64.b64decode(cert)


def starmatch_to_regex(pattern: str) -> Pattern:

    if isinstance(pattern, re.Pattern):
        return pattern
    
    options = re.DOTALL
    # check if we are case-sensitive
    if pattern.startswith("(?-i)"):
        pattern = pattern[5:]
    else:
        options |= re.IGNORECASE
    i, n = 0, len(pattern)
    res = []
    while i < n:
        c = pattern[i]
        i = i + 1
        if c == "*":
            res.append(".*")
        else:
            res.append(re.escape(c))
    return re.compile(r"(?:%s)\Z" % "".join(res), options)


def nested_key(d: dict, *args):
    """
    Traverses a dictionary for nested keys. Returns `None` if the at any point
    in the traversal a key cannot be found.

    Example:

        >>> from atatus.utils import nested_key
        >>> d = {"a": {"b": {"c": 0}}}
        >>> nested_key(d, "a", "b", "c")
        0
        >>> nested_key(d, "a", "b", "d")
        None
    """
    for arg in args:
        try:
            d = d[arg]
        except (TypeError, KeyError):
            d = None
            break
    return d


def getfqdn() -> str:
    """
    socket.getfqdn() has some issues. For one, it's slow (may do a DNS lookup).
    For another, it can return `localhost.localdomain`[1], which is less useful
    than socket.gethostname().

    This function handles the fallbacks and also ensures we don't try to lookup
    the fqdn more than once.

    [1]: https://stackoverflow.com/a/43330159
    """
    global fqdn
    if not fqdn:
        fqdn = socket.getfqdn()
        if fqdn == "localhost.localdomain":
            fqdn = socket.gethostname()
        if not fqdn:
            fqdn = os.environ.get("HOSTNAME")
        if not fqdn:
            fqdn = os.environ.get("HOST")
        if fqdn is None:
            fqdn = ""
        fqdn = fqdn.lower().strip()
    return fqdn



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from __future__ import absolute_import

import contextvars
from typing import TYPE_CHECKING

from atatus.context.base import BaseContext

if TYPE_CHECKING:
    import atatus.traces


class ContextVarsContext(BaseContext):
    atatus_transaction_var = contextvars.ContextVar("atatus_transaction_var")
    atatus_spans_var = contextvars.ContextVar("atatus_spans_var", default=())

    def get_transaction(self, clear: bool = False) -> "atatus.traces.Transaction":
        """
        Get the transaction for the current execution context

        If clear=True, also set the transaction to None for the current
        execution context.
        """
        try:
            transaction = self.atatus_transaction_var.get()
            if clear:
                self.set_transaction(None)
            return transaction
        except LookupError:
            return None

    def set_transaction(self, transaction: "atatus.traces.Transaction") -> None:
        """
        Set the transaction for the current execution context
        """
        self.atatus_transaction_var.set(transaction)

    def get_span(self) -> "atatus.traces.Span":
        """
        Get the active span for the current execution context.
        """
        spans = self.atatus_spans_var.get()
        return spans[-1] if spans else None

    def set_span(self, span: "atatus.traces.Span") -> None:
        """
        Set the active span for the current execution context.

        The previously-activated span will be saved to be re-activated later.
        """
        spans = self.atatus_spans_var.get()
        self.atatus_spans_var.set(spans + (span,))

    def unset_span(self, clear_all: bool = False) -> "atatus.traces.Span":
        """
        De-activate the current span. If a span was previously active, it will
        become active again.

        Returns the deactivated span.

        If clear_all=True, all spans will be cleared and no span will be active.
        """
        spans = self.atatus_spans_var.get()
        span = None
        if spans:
            span = spans[-1]
            if clear_all:
                self.atatus_spans_var.set(())
            else:
                self.atatus_spans_var.set(spans[0:-1])
        return span


execution_context = ContextVarsContext()



#  BSD 3-Clause License
#
#  Copyright (c) 2012, the Sentry Team, see AUTHORS for more details
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE


from __future__ import absolute_import

import logging
import sys
import traceback
import warnings

import wrapt

from atatus import get_client
from atatus.base import Client
from atatus.traces import execution_context
from atatus.utils.stacks import iter_stack_frames


class LoggingHandler(logging.Handler):
    def __init__(self, *args, **kwargs) -> None:
        warnings.warn(
            "The LoggingHandler is deprecated and will be removed in v7.0 of "
            "the agent. Please use `log_ecs_reformatting` and ship the logs "
            "with Elastic Agent or Filebeat instead. "
            "https://www.elastic.co/guide/en/apm/agent/python/current/logs.html",
            DeprecationWarning,
        )
        self.client = None
        if "client" in kwargs:
            self.client = kwargs.pop("client")
        elif len(args) > 0:
            arg = args[0]
            if isinstance(arg, Client):
                self.client = arg

        if not self.client:
            client_cls = kwargs.pop("client_cls", None)
            if client_cls:
                self.client = client_cls(*args, **kwargs)
            else:
                warnings.warn(
                    "LoggingHandler requires a Client instance. No Client was received.",
                    DeprecationWarning,
                )
                self.client = Client(*args, **kwargs)
        logging.Handler.__init__(self, level=kwargs.get("level", logging.NOTSET))

    def emit(self, record):
        self.format(record)

        # Avoid typical config issues by overriding loggers behavior
        if record.name.startswith(("atatus.errors",)):
            sys.stderr.write(record.getMessage() + "\n")
            return

        try:
            return self._emit(record)
        except Exception:
            sys.stderr.write("Top level Atatus exception caught - failed creating log record.\n")
            sys.stderr.write(record.getMessage() + "\n")
            sys.stderr.write(traceback.format_exc() + "\n")

            try:
                self.client.capture("Exception")
            except Exception:
                pass

    def _emit(self, record, **kwargs):
        data = {}

        for k, v in record.__dict__.items():
            if "." not in k and k not in ("culprit",):
                continue
            data[k] = v

        stack = getattr(record, "stack", None)
        if stack is True:
            stack = iter_stack_frames(config=self.client.config)

        if stack:
            frames = []
            started = False
            last_mod = ""
            for item in stack:
                if isinstance(item, (list, tuple)):
                    frame, lineno = item
                else:
                    frame, lineno = item, item.f_lineno

                if not started:
                    f_globals = getattr(frame, "f_globals", {})
                    module_name = f_globals.get("__name__", "")
                    if last_mod.startswith("logging") and not module_name.startswith("logging"):
                        started = True
                    else:
                        last_mod = module_name
                        continue
                frames.append((frame, lineno))
            stack = frames

        custom = getattr(record, "data", {})
        # Add in all of the data from the record that we aren't already capturing
        for k in record.__dict__.keys():
            if k in (
                "stack",
                "name",
                "args",
                "msg",
                "levelno",
                "exc_text",
                "exc_info",
                "data",
                "created",
                "levelname",
                "msecs",
                "relativeCreated",
            ):
                continue
            if k.startswith("_"):
                continue
            custom[k] = record.__dict__[k]

        # If there's no exception being processed,
        # exc_info may be a 3-tuple of None
        # http://docs.python.org/library/sys.html#sys.exc_info
        if record.exc_info and all(record.exc_info):
            handler = self.client.get_handler("atatus.events.Exception")
            exception = handler.capture(self.client, exc_info=record.exc_info)
        else:
            exception = None

        return self.client.capture(
            "Message",
            param_message={"message": str(record.msg), "params": record.args},
            stack=stack,
            custom=custom,
            exception=exception,
            level=record.levelno,
            logger_name=record.name,
            **kwargs,
        )


class LoggingFilter(logging.Filter):
    """
    This filter doesn't actually do any "filtering" -- rather, it just adds
    three new attributes to any "filtered" LogRecord objects:

    * atatus_transaction_id
    * atatus_trace_id
    * atatus_span_id
    * atatus_service_name

    These attributes can then be incorporated into your handlers and formatters,
    so that you can tie log messages to transactions in elasticsearch.

    This filter also adds these fields to a dictionary attribute,
    `atatus_labels`, using the official tracing fields names as documented
    here: https://www.elastic.co/guide/en/ecs/current/ecs-tracing.html

    Note that if you're using Python 3.2+, by default we will add a
    LogRecordFactory to your root logger which will add these attributes
    automatically.
    """

    def __init__(self, name=""):
        super().__init__(name=name)
        warnings.warn(
            "The LoggingFilter is deprecated and will be removed in v7.0 of "
            "the agent. On Python 3.2+, by default we add a LogRecordFactory to "
            "your root logger automatically"
            "https://www.elastic.co/guide/en/apm/agent/python/current/logs.html",
            DeprecationWarning,
        )

    def filter(self, record):
        """
        Add atatus attributes to `record`.
        """
        _add_attributes_to_log_record(record)
        return True


@wrapt.decorator
def log_record_factory(wrapped, instance, args, kwargs):
    """
    Decorator, designed to wrap the python log record factory (fetched by
    logging.getLogRecordFactory), adding the same custom attributes as in
    the LoggingFilter provided above.

    :return:
        LogRecord object, with custom attributes for APM tracing fields
    """
    record = wrapped(*args, **kwargs)
    return _add_attributes_to_log_record(record)


def _add_attributes_to_log_record(record):
    """
    Add custom attributes for APM tracing fields to a LogRecord object

    :param record: LogRecord object
    :return: Updated LogRecord object with new APM tracing fields
    """
    transaction = execution_context.get_transaction()

    transaction_id = transaction.id if transaction else None
    record.atatus_transaction_id = transaction_id

    trace_id = transaction.trace_parent.trace_id if transaction and transaction.trace_parent else None
    record.atatus_trace_id = trace_id

    span = execution_context.get_span()
    span_id = span.id if span else None
    record.atatus_span_id = span_id

    # client = get_client()
    # service_name = client.config.service_name if client else None
    # record.atatus_service_name = service_name

    # service_environment = client.config.environment if client else None
    # record.atatus_service_environment = service_environment

    record.atatus_labels = {
        "transaction.id": transaction_id,
        "trace.id": trace_id,
        "span.id": span_id,
        # "service.name": service_name,
        # "service.environment": service_environment,
    }

    return record


class Formatter(logging.Formatter):
    """
    Custom formatter to automatically append the atatus format string,
    as well as ensure that LogRecord objects actually have the required fields
    (so as to avoid errors which could occur for logs before we override the
    LogRecordFactory):

        formatstring = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        formatstring = formatstring + " | atatus " \
                                      "transaction.id=%(atatus_transaction_id)s " \
                                      "trace.id=%(atatus_trace_id)s " \
                                      "span.id=%(atatus_span_id)s"
    """

    def __init__(self, fmt=None, datefmt=None, style="%") -> None:
        if fmt is None:
            fmt = "%(message)s"
        super(Formatter, self).__init__(fmt=fmt, datefmt=datefmt, style=style)

    def format(self, record):
        formatted_record = super(Formatter, self).format(record=record)

        get = getattr

        tx_id = get(record, "atatus_transaction_id", None)
        if not tx_id:
            return formatted_record

        parts = [ formatted_record, " | atatus " ]
        # parts.append(f"transaction.id={tx_id} ")

        trace_id = get(record, "atatus_trace_id", None)
        if trace_id:
            parts.append(f"trace.id={trace_id} ")

        span_id = get(record, "atatus_span_id", None)
        if span_id:
            parts.append(f"span.id={span_id} ")

        service_name = get(record, "atatus_service_name", None)
        if service_name:
            parts.append(f"service.name={service_name} ")

        service_environment = get(record, "atatus_service_environment", None)
        if service_environment:
            parts.append(f"service.environment={service_environment} ")

        formatted_record = "".join(parts)

        return formatted_record

    def formatTime(self, record, datefmt=None):
        if not hasattr(record, "atatus_transaction_id"):
            record.atatus_transaction_id = None
            record.atatus_trace_id = None
            record.atatus_span_id = None
            record.atatus_service_name = None
            record.atatus_service_environment = None
        return super(Formatter, self).formatTime(record=record, datefmt=datefmt)



#  BSD 3-Clause License
#
#  Copyright (c) 2022, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import inspect

import grpc

import atatus
from atatus.contrib.grpc.server_interceptor import _ServicerContextWrapper, get_trace_parent


class _AsyncServerInterceptor(grpc.aio.ServerInterceptor):
    async def intercept_service(self, continuation, handler_call_details):
        def wrap_unary_unary(behavior):
            async def _interceptor(request, context):
                tp = get_trace_parent(handler_call_details)
                client = atatus.get_client()
                transaction = client.begin_transaction("request", trace_parent=tp)
                try:
                    result = behavior(request, _ServicerContextWrapper(context, transaction))

                    # This is so we can support both sync and async rpc functions
                    if inspect.isawaitable(result):
                        result = await result

                    if transaction and not transaction.outcome:
                        transaction.set_success()
                    return result
                except Exception:
                    if transaction:
                        transaction.set_failure()
                    client.capture_exception(handled=False)
                    raise
                finally:
                    client.end_transaction(name=handler_call_details.method)

            return _interceptor

        handler = await continuation(handler_call_details)
        if handler.request_streaming or handler.response_streaming:
            return handler

        return grpc.unary_unary_rpc_method_handler(
            wrap_unary_unary(handler.unary_unary),
            request_deserializer=handler.request_deserializer,
            response_serializer=handler.response_serializer,
        )



#  BSD 3-Clause License
#
#  Copyright (c) 2022, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Optional

import grpc
from grpc._interceptor import _ClientCallDetails

import atatus
from atatus.conf import constants
from atatus.traces import DroppedSpan, SpanType
from atatus.utils import default_ports


class _ClientInterceptor(
    grpc.UnaryUnaryClientInterceptor,
    # grpc.UnaryStreamClientInterceptor,
    # grpc.StreamUnaryClientInterceptor,
    # grpc.StreamStreamClientInterceptor,
):
    def __init__(self, host: Optional[str], port: Optional[str], secure: bool) -> None:
        self.host: str = host
        self.port: str = port
        self.secure: bool = secure
        schema = "https" if secure else "http"
        resource = f"{schema}://{host}"
        if port and int(port) != default_ports[schema]:
            resource += f":{port}"

        self._context = {
            "http": {
                "url": resource,
            },
            "destination": {
                "address": host,
                "port": port,
            },
        }

    def intercept_unary_unary(self, continuation, client_call_details, request):
        """Intercepts a unary-unary invocation asynchronously.

        Args:
          continuation: A function that proceeds with the invocation by
            executing the next interceptor in chain or invoking the
            actual RPC on the underlying Channel. It is the interceptor's
            responsibility to call it if it decides to move the RPC forward.
            The interceptor can use
            `response_future = continuation(client_call_details, request)`
            to continue with the RPC. `continuation` returns an object that is
            both a Call for the RPC and a Future. In the event of RPC
            completion, the return Call-Future's result value will be
            the response message of the RPC. Should the event terminate
            with non-OK status, the returned Call-Future's exception value
            will be an RpcError.
          client_call_details: A ClientCallDetails object describing the
            outgoing RPC.
          request: The request value for the RPC.

        Returns:
            An object that is both a Call for the RPC and a Future.
            In the event of RPC completion, the return Call-Future's
            result value will be the response message of the RPC.
            Should the event terminate with non-OK status, the returned
            Call-Future's exception value will be an RpcError.
        """
        with atatus.capture_span(
            client_call_details.method, span_type="external", span_subtype="grpc", extra=self._context.copy(), leaf=True
        ) as span:
            if not span:
                return continuation(client_call_details, request)
            client_call_details = self.attach_traceparent(client_call_details, span)
            try:
                response = continuation(client_call_details, request)
            except grpc.RpcError:
                span.set_failure()
                raise

            return response

    # TODO: instrument other types of requests once the spec is ready

    # def intercept_unary_stream(self, continuation, client_call_details,
    #                            request):
    #     """Intercepts a unary-stream invocation.
    #
    #     Args:
    #       continuation: A function that proceeds with the invocation by
    #         executing the next interceptor in chain or invoking the
    #         actual RPC on the underlying Channel. It is the interceptor's
    #         responsibility to call it if it decides to move the RPC forward.
    #         The interceptor can use
    #         `response_iterator = continuation(client_call_details, request)`
    #         to continue with the RPC. `continuation` returns an object that is
    #         both a Call for the RPC and an iterator for response values.
    #         Drawing response values from the returned Call-iterator may
    #         raise RpcError indicating termination of the RPC with non-OK
    #         status.
    #       client_call_details: A ClientCallDetails object describing the
    #         outgoing RPC.
    #       request: The request value for the RPC.
    #
    #     Returns:
    #         An object that is both a Call for the RPC and an iterator of
    #         response values. Drawing response values from the returned
    #         Call-iterator may raise RpcError indicating termination of
    #         the RPC with non-OK status. This object *should* also fulfill the
    #         Future interface, though it may not.
    #     """
    #     response_iterator = continuation(client_call_details, request)
    #     return response_iterator
    #
    # def intercept_stream_unary(self, continuation, client_call_details,
    #                            request_iterator):
    #     """Intercepts a stream-unary invocation asynchronously.
    #
    #     Args:
    #       continuation: A function that proceeds with the invocation by
    #         executing the next interceptor in chain or invoking the
    #         actual RPC on the underlying Channel. It is the interceptor's
    #         responsibility to call it if it decides to move the RPC forward.
    #         The interceptor can use
    #         `response_future = continuation(client_call_details, request_iterator)`
    #         to continue with the RPC. `continuation` returns an object that is
    #         both a Call for the RPC and a Future. In the event of RPC completion,
    #         the return Call-Future's result value will be the response message
    #         of the RPC. Should the event terminate with non-OK status, the
    #         returned Call-Future's exception value will be an RpcError.
    #       client_call_details: A ClientCallDetails object describing the
    #         outgoing RPC.
    #       request_iterator: An iterator that yields request values for the RPC.
    #
    #     Returns:
    #       An object that is both a Call for the RPC and a Future.
    #       In the event of RPC completion, the return Call-Future's
    #       result value will be the response message of the RPC.
    #       Should the event terminate with non-OK status, the returned
    #       Call-Future's exception value will be an RpcError.
    #     """
    #
    # def intercept_stream_stream(self, continuation, client_call_details,
    #                             request_iterator):
    #     """Intercepts a stream-stream invocation.
    #
    #     Args:
    #       continuation: A function that proceeds with the invocation by
    #         executing the next interceptor in chain or invoking the
    #         actual RPC on the underlying Channel. It is the interceptor's
    #         responsibility to call it if it decides to move the RPC forward.
    #         The interceptor can use
    #         `response_iterator = continuation(client_call_details, request_iterator)`
    #         to continue with the RPC. `continuation` returns an object that is
    #         both a Call for the RPC and an iterator for response values.
    #         Drawing response values from the returned Call-iterator may
    #         raise RpcError indicating termination of the RPC with non-OK
    #         status.
    #       client_call_details: A ClientCallDetails object describing the
    #         outgoing RPC.
    #       request_iterator: An iterator that yields request values for the RPC.
    #
    #     Returns:
    #       An object that is both a Call for the RPC and an iterator of
    #       response values. Drawing response values from the returned
    #       Call-iterator may raise RpcError indicating termination of
    #       the RPC with non-OK status. This object *should* also fulfill the
    #       Future interface, though it may not.
    #     """

    def attach_traceparent(self, client_call_details: _ClientCallDetails, span: SpanType) -> _ClientCallDetails:
        if isinstance(span, DroppedSpan) or not span.transaction:
            return client_call_details
        meta = list(client_call_details.metadata) if client_call_details.metadata else []
        if constants.TRACEPARENT_HEADER_NAME not in meta:
            traceparent = span.transaction.trace_parent.copy_from(span_id=span.id)
            meta.extend(
                (
                    (constants.TRACEPARENT_HEADER_NAME, traceparent.to_string()),
                    (constants.TRACESTATE_HEADER_NAME, traceparent.tracestate),
                )
            )
        return client_call_details._replace(metadata=meta)



#  BSD 3-Clause License
#
#  Copyright (c) 2023, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import threading
from http.cookies import SimpleCookie
from typing import Dict, Optional, TypeVar

import azure.functions as func
from azure.functions.extension import AppExtensionBase

import atatus
from atatus.base import Client
from atatus.conf import constants
from atatus.utils import get_url_dict
from atatus.utils.disttracing import TraceParent
from atatus.utils.logging import get_logger

logger = get_logger("atatus.serverless")

_AnnotatedFunctionT = TypeVar("_AnnotatedFunctionT")

_cold_start_lock = threading.Lock()
COLD_START = True


class AzureFunctionsClient(Client):
    def get_service_info(self):
        service_info = super().get_service_info()
        service_info["framework"] = {
            "name": "Azure Functions",
            "version": os.environ.get("FUNCTIONS_EXTENSION_VERSION"),
        }
        service_info["runtime"] = {
            "name": os.environ.get("FUNCTIONS_WORKER_RUNTIME"),
            "version": os.environ.get("FUNCTIONS_WORKER_RUNTIME_VERSION"),
        }
        service_info["node"] = {"configured_name": os.environ.get("WEBSITE_INSTANCE_ID")}
        return service_info

    def get_cloud_info(self):
        cloud_info = super().get_cloud_info()
        cloud_info.update(
            {"provider": "azure", "region": os.environ.get("REGION_NAME"), "service": {"name": "functions"}}
        )
        account_id = get_account_id()
        if account_id:
            cloud_info["account"] = {"id": account_id}
        if "WEBSITE_SITE_NAME" in os.environ:
            cloud_info["instance"] = {"name": os.environ["WEBSITE_SITE_NAME"]}
        if "WEBSITE_RESOURCE_GROUP" in os.environ:
            cloud_info["project"] = {"name": os.environ["WEBSITE_RESOURCE_GROUP"]}
        return cloud_info


class AtatusExtension(AppExtensionBase):
    client = None

    @classmethod
    def init(cls) -> None:
        """The function will be executed when the extension is loaded.
        Happens when Azure Functions customers import the extension module.
        """
        atatus.instrument()

    @classmethod
    def configure(cls, client_class=AzureFunctionsClient, **kwargs) -> None:
        client = atatus.get_client()
        if not client:
            kwargs["metrics_interval"] = "0ms"
            kwargs["breakdown_metrics"] = "false"
            if "metric_sets" not in kwargs and "ATATUS_METRICS_SETS" not in os.environ:
                # Allow users to override metrics sets
                kwargs["metrics_sets"] = []
            kwargs["central_config"] = "false"
            kwargs["cloud_provider"] = "none"
            kwargs["framework_name"] = "Azure Functions"
            if (
                "service_name" not in kwargs
                and "ATATUS_SERVICE_NAME" not in os.environ
                and "WEBSITE_SITE_NAME" in os.environ
            ):
                kwargs["service_name"] = os.environ["WEBSITE_SITE_NAME"]
            if (
                "environment" not in kwargs
                and "ATATUS_ENVIRONMENT" not in os.environ
                and "AZURE_FUNCTIONS_ENVIRONMENT" in os.environ
            ):
                kwargs["environment"] = os.environ["AZURE_FUNCTIONS_ENVIRONMENT"]
            client = AzureFunctionsClient(**kwargs)
        cls.client = client

    @classmethod
    def pre_invocation_app_level(cls, logger, context, func_args: Dict[str, object] = None, *args, **kwargs) -> None:
        """
        This must be implemented as a @staticmethod. It will be called right
        before a customer's function is being executed.
        """
        client = cls.client
        if not client:
            return
        global COLD_START
        with _cold_start_lock:
            cold_start, COLD_START = COLD_START, False
        tp: Optional[TraceParent] = None
        transaction_type = "request"
        http_request_data = None
        trigger_type = None
        if func_args:
            for arg in func_args.values():
                if isinstance(arg, func.HttpRequest):
                    tp = TraceParent.from_headers(arg.headers) if arg.headers else None
                    transaction_type = "request"
                    http_request_data = lambda: get_data_from_request(
                        arg, client.config.capture_headers, client.config.capture_body in ("transactions", "all")
                    )
                    trigger_type = "request"
                    break
                if isinstance(arg, func.TimerRequest):
                    transaction_type = "timer"
                    trigger_type = "timer"
                    break

        cls.client.begin_transaction(transaction_type, trace_parent=tp)
        if http_request_data:
            atatus.set_context(http_request_data, "request")
        atatus.set_context(lambda: get_faas_data(context, cold_start, trigger_type), "faas")

    @classmethod
    def post_invocation_app_level(
        cls, logger, context, func_args: Dict[str, object] = None, func_ret=Optional[object], *args, **kwargs
    ) -> None:
        client = cls.client
        if isinstance(func_ret, func.HttpResponse):
            atatus.set_context(lambda: get_data_from_response(func_ret, client.config.capture_headers), "response")
            atatus.set_transaction_outcome(http_status_code=func_ret.status_code)
            atatus.set_transaction_result(f"HTTP {func_ret.status_code // 100}xx")
        client.end_transaction(context.function_name)


def get_data_from_request(request: func.HttpRequest, capture_headers=False, capture_body=False):
    result = {
        "method": request.method,
    }
    if "Cookie" in request.headers:
        cookie = SimpleCookie()
        cookie.load(request.headers["Cookie"])
        result["cookies"] = {k: v.value for k, v in cookie.items()}
    if capture_headers and request.headers:
        result["headers"] = dict(request.headers)
    if capture_body and request.method in constants.HTTP_WITH_BODY:
        result["body"] = request.get_body()
    result["url"] = get_url_dict(request.url)
    return result


def get_data_from_response(response: func.HttpResponse, capture_headers: bool = False):
    result = {
        "status_code": response.status_code,
    }
    if capture_headers:
        result["headers"] = dict(response.headers)
    return result


def get_faas_data(context: func.Context, cold_start: bool, trigger_type: Optional[str]) -> dict:
    account_id = get_account_id()
    resource_group = os.environ.get("WEBSITE_RESOURCE_GROUP", None)
    app_name = os.environ.get("WEBSITE_SITE_NAME", None)
    function_name = context.function_name
    data = {
        "coldstart": cold_start,
        "execution": context.invocation_id,
    }
    if trigger_type:
        data["trigger"] = {"type": trigger_type}
    if account_id and resource_group and app_name and function_name:
        data["id"] = (
            f"/subscriptions/{account_id}/resourceGroups/{resource_group}/providers/Microsoft.Web/sites/{app_name}/"
            + f"functions/{function_name}"
        )
    if app_name and function_name:
        data["name"] = f"{app_name}/{function_name}"
    return data


def get_account_id() -> Optional[str]:
    return os.environ["WEBSITE_OWNER_NAME"].split("+", 1)[0] if "WEBSITE_OWNER_NAME" in os.environ else None



#  BSD 3-Clause License
#
#  Copyright (c) 2012, the Sentry Team, see AUTHORS for more details
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE


from __future__ import absolute_import

import sys
import typing as t

from sanic import Sanic
from sanic.request import Request
from sanic.response import HTTPResponse

from atatus import label
from atatus import set_context as atatus_context
from atatus import (
    set_custom_context,
    set_transaction_name,
    set_transaction_outcome,
    set_transaction_result,
    set_user_context,
)
from atatus.base import Client
from atatus.conf import constants
from atatus.contrib.asyncio.traces import set_context
from atatus.contrib.sanic.patch import AtatusPatchedErrorHandler
from atatus.contrib.sanic.sanic_types import (
    AllMiddlewareGroup,
    APMConfigType,
    CustomContextCallbackType,
    CustomInfoType,
    ExtendableMiddlewareGroup,
    LabelInfoCallbackType,
    TransactionNameCallbackType,
    UserInfoCallbackType,
)
from atatus.contrib.sanic.utils import SanicAPMConfig, get_request_info, get_response_info, make_client
from atatus.instrumentation.control import instrument
from atatus.utils.disttracing import TraceParent
from atatus.utils.logging import get_logger


class Atatus:
    """
    Sanic App Middleware for Atatus Capturing

    >>> app = Sanic(name="atatus-sample")

    Pass the Sanic app and let the configuration be derived from it::

    >>> apm = Atatus(app=app)

    Configure the APM Client Using Custom Configurations::

    >>> apm = Atatus(app=app, config={
        "SERVICE_NAME": "atatus-sample",
        "SERVICE_VERSION": "v1.2.0",
        "SERVER_URL": "http://eapm-server.somdomain.com:443",
        "SECRET_TOKEN": "supersecrettokenstuff",
    })

    Pass a pre-build Client instance to the APM Middleware::

    >>> apm = Atatus(app=app, client=Client())

    Pass arbitrary Server name and token to the client while initializing::

    >>> apm = Atatus(app=app, service_name="atatus-sample", secret_token="supersecretthing")

    Capture an Exception::

    >>> try:
    >>>     1 / 0
    >>> except ZeroDivisionError:
    >>>     apm.capture_exception()

    Capture generic message::

    >>> apm.capture_message("Some Nice message to be captured")
    """

    def __init__(
        self,
        app: Sanic,
        client: t.Optional[Client] = None,
        client_cls: t.Type[Client] = Client,
        config: APMConfigType = None,
        transaction_name_callback: TransactionNameCallbackType = None,
        user_context_callback: UserInfoCallbackType = None,
        custom_context_callback: CustomContextCallbackType = None,
        label_info_callback: LabelInfoCallbackType = None,
        **defaults,
    ) -> None:
        """
        Initialize an instance of the Atatus client that will be used to configure the reset of the Application
        middleware

        :param app: An instance of Sanic app server
        :param client: An instance of Client if you want to leverage a custom APM client instance pre-created
        :param client_cls: Base Instance of the Elastic Client to be used to setup the APM Middleware
        :param config: Configuration values to be used for setting up the Elastic Client. This includes the APM server
        :param transaction_name_callback: Callback method used to extract the transaction name. If nothing is provided
            it will fallback to the default implementation provided by the middleware extension
        :param user_context_callback: Callback method used to extract the user context information. Will be ignored
            if one is not provided by the users while creating an instance of the Atatus client
        :param custom_context_callback: Callback method used to generate custom context information for the transaction
        :param label_info_callback: Callback method used to generate custom labels/tags for the current transaction
        :param defaults: Default configuration values to be used for settings up the APM client
        """
        self._app = app  # type: Sanic
        self._client_cls = client_cls  # type: type
        self._client = client  # type: t.Union[None, Client]
        self._skip_init_middleware = defaults.pop("skip_init_middleware", False)  # type: bool
        self._skip_init_exception_handler = defaults.pop("skip_init_exception_handler", False)  # type: bool
        self._transaction_name_callback = transaction_name_callback  # type: TransactionNameCallbackType
        self._user_context_callback = user_context_callback  # type: UserInfoCallbackType
        self._custom_context_callback = custom_context_callback  # type: CustomContextCallbackType
        self._label_info_callback = label_info_callback  # type: LabelInfoCallbackType
        self._logger = get_logger("atatus.errors.client")
        self._client_config = {}  # type: t.Dict[str, str]
        self._setup_client_config(config=config)
        self._init_app()

    async def capture_exception(self, exc_info=None, handled=True, **kwargs):
        """
        Capture a generic exception and traceback to be reported to the APM
        :param exc_info: Exc info extracted from the traceback for the current exception
        :param handled: Boolean indicator for if the exception is handled.
        :param kwargs: additional context to be passed to the API client for capturing exception related information
        :return: None
        """
        assert self._client, "capture_exception called before application configuration is initialized"
        return self._client.capture_exception(exc_info=exc_info, handled=handled, **kwargs)

    async def capture_message(self, message=None, param_message=None, **kwargs):
        """
        Capture a generic message for the APM Client
        :param message: Message information to be captured
        :param param_message:
        :param kwargs: additional context to be passed to the API client for capturing exception related information
        :return:
        """
        assert self._client, "capture_message called before application configuration is initialized"
        return self._client.capture_message(message=message, param_message=param_message, **kwargs)

    def _setup_client_config(self, config: APMConfigType = None) -> None:
        app_based_config = SanicAPMConfig(self._app)
        if dict(app_based_config):
            self._client_config = dict(app_based_config)

        if config:
            self._client_config.update(config)

    # noinspection PyBroadException,PyUnresolvedReferences
    def _init_app(self) -> None:
        """
        Initialize all the required middleware and other application infrastructure that will perform the necessary
        capture of the APM instrumentation artifacts
        :return: None
        """
        if not self._client:
            self._client = make_client(config=self._client_config, client_cls=self._client_cls, **self._client_config)

        if not self._skip_init_exception_handler:
            self._setup_exception_manager()

        if self._client.config.instrument and self._client.config.enabled:
            instrument()
            try:
                from atatus.contrib.celery import register_instrumentation

                register_instrumentation(client=self._client)
            except ImportError:
                self._logger.debug(
                    "Failed to setup instrumentation. "
                    "Please install requirements for atatus.contrib.celery if instrumentation is required"
                )
                pass

        if not self._skip_init_middleware:
            self._setup_request_handler(entity=self._app)

    # noinspection PyMethodMayBeStatic,PyBroadException
    def _default_transaction_name_generator(self, request: Request) -> str:
        """
        Method used to extract the default transaction name. This is generated by joining the HTTP method and the
        URL path used for invoking the API handler
        :param request: Sanic HTTP Request object
        :return: string containing the Transaction name
        """
        url_template = request.path
        # Sanic's new router puts this into the request itself so that it can be accessed easily
        # On Exception with `NotFound` with new Sanic Router, the `route` object will be None
        # This check is to enforce that limitation
        if hasattr(request, "route") and request.route:
            url_template = request.route.path
            url_template = f"/{url_template}" if not url_template.startswith("/") else url_template
        else:
            # Let us fallback to using old router model to extract the info
            try:
                _, _, _, url_template, _, _ = self._app.router.get(request=request)
            except Exception:
                pass

        return f"{request.method} {url_template}"

    # noinspection PyMethodMayBeStatic
    async def _setup_default_custom_context(self, request: Request) -> CustomInfoType:
        return request.match_info

    def setup_middleware(self, entity: ExtendableMiddlewareGroup) -> None:
        """
        Adhoc registration of the middlewares for Blueprint and BlueprintGroup if you don't want to instrument
        your entire application. Only part of it can be done.
        :param entity: Blueprint or BlueprintGroup Kind of resource
        :return: None
        """
        self._setup_request_handler(entity=entity)

    def _setup_request_handler(self, entity: AllMiddlewareGroup) -> None:
        """
        This method is used to setup a series of Sanic Application level middleware so that they can be applied to all
        the routes being registered under the app easily.

        :param entity: entity: Sanic APP or Blueprint or BlueprintGroup Kind of resource
        :return: None
        """

        @entity.middleware("request")
        async def _instrument_request(request: Request) -> None:
            if not self._client.should_ignore_url(url=request.path):
                trace_parent = TraceParent.from_headers(headers=request.headers)
                self._client.begin_transaction("request", trace_parent=trace_parent)
                await set_context(
                    lambda: get_request_info(
                        config=self._client.config, request=request, event_type=constants.TRANSACTION
                    ),
                    "request",
                )
                self._setup_transaction_name(request=request)
                if self._user_context_callback:
                    name, email, uid = await self._user_context_callback(request)
                    set_user_context(username=name, email=email, user_id=uid)

                await self._setup_custom_context(request=request)

                if self._label_info_callback:
                    labels = await self._label_info_callback(request)
                    label(**labels)

        # noinspection PyUnusedLocal
        @entity.middleware("response")
        async def _instrument_response(request: Request, response: HTTPResponse) -> None:
            await set_context(
                lambda: get_response_info(
                    config=self._client.config, response=response, event_type=constants.TRANSACTION
                ),
                "response",
            )
            self._setup_transaction_name(request=request)
            result = f"HTTP {response.status // 100}xx"
            set_transaction_result(result=result, override=False)
            set_transaction_outcome(http_status_code=response.status, override=False)
            atatus_context(data={"status_code": response.status}, key="response")
            self._client.end_transaction()

    def _setup_transaction_name(self, request: Request) -> None:
        """
        Method used to setup the transaction name using the provided callback or the default mode
        :param request: Incoming HTTP Request entity
        :return: None
        """
        if self._transaction_name_callback:
            name = self._transaction_name_callback(request)
        else:
            name = self._default_transaction_name_generator(request=request)
        if name:
            set_transaction_name(name, override=False)

    async def _setup_custom_context(self, request: Request) -> None:
        if self._custom_context_callback:
            set_custom_context(data=await self._custom_context_callback(request))
        else:
            set_custom_context(data=await self._setup_default_custom_context(request=request))

    # noinspection PyBroadException,PyProtectedMember
    def _setup_exception_manager(self) -> None:
        """
        Setup global exception handler where all unhandled exception can be caught and tracked to APM server
        :return:
        """

        # noinspection PyUnusedLocal
        async def _handler(request: Request, exception: BaseException) -> None:
            if not self._client:
                return

            self._client.capture_exception(
                exc_info=sys.exc_info(),
                context={
                    "request": await get_request_info(
                        config=self._client.config, request=request, event_type=constants.ERROR
                    ),
                },
                handled=True,
            )
            self._setup_transaction_name(request=request)
            set_transaction_result(result="HTTP 5xx", override=False)
            set_transaction_outcome(outcome=constants.OUTCOME.FAILURE, override=False)
            atatus_context(data={"status_code": 500}, key="response")
            self._client.end_transaction()

        if not isinstance(self._app.error_handler, AtatusPatchedErrorHandler):
            patched_client = AtatusPatchedErrorHandler(current_handler=self._app.error_handler)
            patched_client.setup_apm_handler(apm_handler=_handler)
            self._app.error_handler = patched_client
        else:
            self._app.error_handler.setup_apm_handler(apm_handler=_handler)

        try:
            from atatus.contrib.celery import register_exception_tracking

            register_exception_tracking(client=self._client)
        except ImportError:
            self._logger.debug(
                "Failed to setup Exception tracking. "
                "Please install requirements for atatus.contrib.celery if exception tracking is required"
            )
            pass



import json
from functools import wraps
import os
from typing import (
    Optional,
    TypeVar,
    Callable,
    Any,
    cast,
    ParamSpec,
    Awaitable,
)
import inspect
import warnings
import types
import dataclasses

from opentelemetry import trace
from opentelemetry.trace.status import Status, StatusCode
from opentelemetry import context as context_api
from opentelemetry.semconv_ai import SpanAttributes, TraceloopSpanKindValues

from atatus import get_client
from atatus.contrib.opentelemetry import get_tracer
from atatus.llmobs.instrumentor import set_workflow_name, set_entity_path , get_chained_entity_path

# from traceloop.sdk.tracing.tracing import (
#     TracerWrapper,
# )
class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, dict):
            if "callbacks" in o:
                del o["callbacks"]
                return o
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)

        if hasattr(o, "to_json"):
            return o.to_json()

        if hasattr(o, "json"):
            return o.json()

        if hasattr(o, "__class__"):
            return o.__class__.__name__

        return super().default(o)

P = ParamSpec("P")

R = TypeVar("R")
F = TypeVar("F", bound=Callable[P, R | Awaitable[R]])

def camel_to_snake(s):
    if len(s) <= 1:
        return s.lower()

    return cameltosnake(s[0].lower() + s[1:])

def cameltosnake(camel_string: str) -> str:
    if not camel_string:
        return ""
    elif camel_string[0].isupper():
        return f"_{camel_string[0].lower()}{cameltosnake(camel_string[1:])}"
    else:
        return f"{camel_string[0]}{cameltosnake(camel_string[1:])}"

def _is_json_size_valid(json_str: str) -> bool:
    """Check if JSON string size is less than 1MB"""
    return len(json_str) < 1_000_000


# Async Decorators - Deprecated


def aentity_method(
    name: Optional[str] = None,
    version: Optional[int] = None,
    tlp_span_kind: Optional[TraceloopSpanKindValues] = TraceloopSpanKindValues.TASK,
):
    warnings.warn(
        "DeprecationWarning: The @aentity_method function will be removed in a future version. "
        "Please migrate to @entity_method for both sync and async operations.",
        DeprecationWarning,
        stacklevel=2,
    )

    return entity_method(
        name=name,
        version=version,
        tlp_span_kind=tlp_span_kind,
    )


def aentity_class(
    name: Optional[str],
    version: Optional[int],
    method_name: str,
    tlp_span_kind: Optional[TraceloopSpanKindValues] = TraceloopSpanKindValues.TASK,
):
    warnings.warn(
        "DeprecationWarning: The @aentity_class function will be removed in a future version. "
        "Please migrate to @entity_class for both sync and async operations.",
        DeprecationWarning,
        stacklevel=2,
    )

    return entity_class(
        name=name,
        version=version,
        method_name=method_name,
        tlp_span_kind=tlp_span_kind,
    )


def _handle_generator(span, res):
    # for some reason the SPAN_KEY is not being set in the context of the generator, so we re-set it
    context_api.attach(trace.set_span_in_context(span))
    try:
        for item in res:
            yield item
    except Exception as e:
        span.set_status(Status(StatusCode.ERROR, str(e)))
        span.record_exception(e)
        raise
    finally:
        span.end()
        # Note: we don't detach the context here as this fails in some situations
        # https://github.com/open-telemetry/opentelemetry-python/issues/2606
        # This is not a problem since the context will be detached automatically during garbage collection


async def _ahandle_generator(span, ctx_token, res):
    try:
        async for part in res:
            yield part
    except Exception as e:
        span.set_status(Status(StatusCode.ERROR, str(e)))
        span.record_exception(e)
        raise
    finally:
        span.end()
        context_api.detach(ctx_token)


def _should_send_prompts():
    return (
        os.getenv("TRACELOOP_TRACE_CONTENT") or "true"
    ).lower() == "true" or context_api.get_value("override_enable_content_tracing")


# Unified Decorators : handles both sync and async operations


def _is_async_method(fn):
    # check if co-routine function or async generator( example : using async & yield)
    return inspect.iscoroutinefunction(fn) or inspect.isasyncgenfunction(fn)


def _setup_span(entity_name, tlp_span_kind, version):
    """Sets up the OpenTelemetry span and context"""
    if tlp_span_kind in [
        TraceloopSpanKindValues.WORKFLOW,
        TraceloopSpanKindValues.AGENT,
    ]:
        set_workflow_name(entity_name)
    span_name = f"{entity_name}.{tlp_span_kind.value}"

    tracer = get_tracer(__name__)
    span = tracer.start_span(span_name)
    ctx = trace.set_span_in_context(span)
    ctx_token = context_api.attach(ctx)

    if tlp_span_kind in [
        TraceloopSpanKindValues.TASK,
        TraceloopSpanKindValues.TOOL,
    ]:
        entity_path = get_chained_entity_path(entity_name)
        set_entity_path(entity_path)

    span.set_attribute(SpanAttributes.TRACELOOP_SPAN_KIND, tlp_span_kind.value)
    span.set_attribute(SpanAttributes.TRACELOOP_ENTITY_NAME, entity_name)
    if version:
        span.set_attribute(SpanAttributes.TRACELOOP_ENTITY_VERSION, version)

    return span, ctx, ctx_token


def _handle_span_input(span, args, kwargs, cls=None):
    """Handles entity input logging in JSON for both sync and async functions"""
    try:
        if _should_send_prompts():
            json_input = json.dumps(
                {"args": args, "kwargs": kwargs}, **({"cls": cls} if cls else {})
            )
            if _is_json_size_valid(json_input):
                span.set_attribute(
                    SpanAttributes.TRACELOOP_ENTITY_INPUT,
                    json_input,
                )
    except TypeError:
        get_client().capture_exception()


def _handle_span_output(span, res, cls=None):
    """Handles entity output logging in JSON for both sync and async functions"""
    try:
        if _should_send_prompts():
            json_output = json.dumps(res, **({"cls": cls} if cls else {}))
            if _is_json_size_valid(json_output):
                span.set_attribute(
                    SpanAttributes.TRACELOOP_ENTITY_OUTPUT,
                    json_output,
                )
    except TypeError:
        get_client().capture_exception()


def _cleanup_span(span, ctx_token):
    """End the span process and detach the context token"""
    span.end()
    context_api.detach(ctx_token)


def entity_method(
    name: Optional[str] = None,
    version: Optional[int] = None,
    tlp_span_kind: Optional[TraceloopSpanKindValues] = TraceloopSpanKindValues.TASK,
) -> Callable[[F], F]:
    def decorate(fn: F) -> F:
        is_async = _is_async_method(fn)
        entity_name = name or fn.__qualname__
        if is_async:
            if inspect.isasyncgenfunction(fn):

                @wraps(fn)
                async def async_gen_wrap(*args: Any, **kwargs: Any) -> Any:
                    # if not TracerWrapper.verify_initialized():
                    #     async for item in fn(*args, **kwargs):
                    #         yield item
                    #     return

                    span, ctx, ctx_token = _setup_span(
                        entity_name, tlp_span_kind, version
                    )
                    _handle_span_input(span, args, kwargs, cls=JSONEncoder)
                    async for item in _ahandle_generator(
                        span, ctx_token, fn(*args, **kwargs)
                    ):
                        yield item

                return cast(F, async_gen_wrap)
            else:

                @wraps(fn)
                async def async_wrap(*args: Any, **kwargs: Any) -> Any:
                    # if not TracerWrapper.verify_initialized():
                    #     return await fn(*args, **kwargs)
                    #
                    span, ctx, ctx_token = _setup_span(
                        entity_name, tlp_span_kind, version
                    )
                    _handle_span_input(span, args, kwargs, cls=JSONEncoder)
                    try:
                        res = await fn(*args, **kwargs)
                        _handle_span_output(span, res, cls=JSONEncoder)
                        return res
                    except Exception as e:
                        span.set_status(Status(StatusCode.ERROR, str(e)))
                        span.record_exception(e)
                        raise
                    finally:
                        _cleanup_span(span, ctx_token)

                return cast(F, async_wrap)
        else:

            @wraps(fn)
            def sync_wrap(*args: Any, **kwargs: Any) -> Any:
                # if not TracerWrapper.verify_initialized():
                #     return fn(*args, **kwargs)

                span, ctx, ctx_token = _setup_span(entity_name, tlp_span_kind, version)
                _handle_span_input(span, args, kwargs, cls=JSONEncoder)
                try:
                    res = fn(*args, **kwargs)
                except Exception as e:
                    span.set_status(Status(StatusCode.ERROR, str(e)))
                    span.record_exception(e)
                    _cleanup_span(span, ctx_token)
                    raise

                # span will be ended in the generator
                if isinstance(res, types.GeneratorType):
                    return _handle_generator(span, res)

                _handle_span_output(span, res, cls=JSONEncoder)
                _cleanup_span(span, ctx_token)
                return res

            return cast(F, sync_wrap)

    return decorate


def entity_class(
    name: Optional[str],
    version: Optional[int],
    method_name: str,
    tlp_span_kind: Optional[TraceloopSpanKindValues] = TraceloopSpanKindValues.TASK,
):
    def decorator(cls):
        task_name = name if name else camel_to_snake(cls.__qualname__)
        method = getattr(cls, method_name)
        setattr(
            cls,
            method_name,
            entity_method(name=task_name, version=version, tlp_span_kind=tlp_span_kind)(
                method
            ),
        )
        return cls

    return decorator



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
This module implements WSGI related helpers adapted from ``werkzeug.wsgi``

:copyright: (c) 2010 by the Werkzeug Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from urllib.parse import quote


# `get_headers` comes from `werkzeug.datastructures.EnvironHeaders`
def get_headers(environ):
    """
    Returns only proper HTTP headers.
    """
    for key, value in environ.items():
        key = str(key)
        if key.startswith("HTTP_") and key not in ("HTTP_CONTENT_TYPE", "HTTP_CONTENT_LENGTH"):
            yield key[5:].replace("_", "-").lower(), value
        elif key in ("CONTENT_TYPE", "CONTENT_LENGTH"):
            yield key.replace("_", "-").lower(), value


def get_environ(environ):
    """
    Returns our whitelisted environment variables.
    """
    for key in ("REMOTE_ADDR", "SERVER_NAME", "SERVER_PORT"):
        if key in environ:
            yield key, environ[key]


# `get_host` comes from `werkzeug.wsgi`
def get_host(environ):
    """Return the real host for the given WSGI environment.  This takes care
    of the `X-Forwarded-Host` header.

    :param environ: the WSGI environment to get the host of.
    """
    scheme = environ.get("wsgi.url_scheme")
    if "HTTP_X_FORWARDED_HOST" in environ:
        result = environ["HTTP_X_FORWARDED_HOST"]
    elif "HTTP_HOST" in environ:
        result = environ["HTTP_HOST"]
    else:
        result = environ["SERVER_NAME"]
        if (scheme, str(environ["SERVER_PORT"])) not in (("https", "443"), ("http", "80")):
            result += ":" + environ["SERVER_PORT"]
    if result.endswith(":80") and scheme == "http":
        result = result[:-3]
    elif result.endswith(":443") and scheme == "https":
        result = result[:-4]
    return result


# `get_current_url` comes from `werkzeug.wsgi`
def get_current_url(environ, root_only=False, strip_querystring=False, host_only=False, path_only=False):
    """A handy helper function that recreates the full URL for the current
    request or parts of it.  Here an example:

    >>> from werkzeug import create_environ
    >>> env = create_environ("/?param=foo", "http://localhost/script")
    >>> get_current_url(env)
    'http://localhost/script/?param=foo'
    >>> get_current_url(env, root_only=True)
    'http://localhost/script/'
    >>> get_current_url(env, host_only=True)
    'http://localhost/'
    >>> get_current_url(env, strip_querystring=True)
    'http://localhost/script/'

    :param environ: the WSGI environment to get the current URL from.
    :param root_only: set `True` if you only want the root URL.
    :param strip_querystring: set to `True` if you don't want the querystring.
    :param host_only: set to `True` if the host URL should be returned.
    :param path_only: set to `True` if only the path should be returned.
    """
    if path_only:
        tmp = []
    else:
        tmp = [environ["wsgi.url_scheme"], "://", get_host(environ)]
    cat = tmp.append
    if host_only:
        return "".join(tmp) + "/"
    cat(quote(environ.get("SCRIPT_NAME", "").rstrip("/")))
    if root_only:
        cat("/")
    else:
        cat(quote("/" + environ.get("PATH_INFO", "").lstrip("/")))
        if not strip_querystring:
            qs = environ.get("QUERY_STRING")
            if qs:
                cat("?" + qs)
    return "".join(tmp)
